@@ -20,7 +20,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
-    $Id: IDL.h,v 1.76 1998/12/05 19:15:21 andrewtv Exp $
+    $Id: IDL.h,v 1.80 1998/12/16 21:46:36 andrewtv Exp $
 
 ***************************************************************************/
 #ifndef __IDL_H
@@ -56,6 +56,12 @@ extern "C" {
 #define IDLF_DECLSPEC_EXIST		(1UL << 0)
 #define IDLF_DECLSPEC_INHIBIT		(1UL << 1)
 
+/* output flags */
+#define IDLF_OUTPUT_NO_NEWLINES		(1UL << 0)
+#define IDLF_OUTPUT_NO_QUALIFY_IDENTS	(1UL << 1)
+#define IDLF_OUTPUT_PROPERTIES		(1UL << 2)
+#define IDLF_OUTPUT_CODEFRAGS		(1UL << 3)
+
 #ifdef _WIN32
 #  define IDL_EXPORT			__declspec (dllexport)
 #  define IDL_IMPORT			__declspec (dllimport)
@@ -638,6 +644,7 @@ typedef int		(*IDL_msg_callback)		(int level,
 							 const char *message);
 
 typedef gboolean	(*IDL_tree_func)		(IDL_tree p,
+							 IDL_tree scope,
 							 gpointer user_data);
 	
 extern IDL_tree		IDL_check_type_cast		(const IDL_tree var,
@@ -690,6 +697,8 @@ extern IDL_tree		IDL_get_parent_node		(IDL_tree p,
 							 IDL_tree_type type,
 							 int *scope_levels);
 
+extern IDL_tree		IDL_tree_get_scope		(IDL_tree p);
+
 extern int		IDL_tree_get_node_info		(IDL_tree tree,
 							 char **who,
 							 char **what);
@@ -713,12 +722,23 @@ extern void		IDL_tree_property_set		(IDL_tree tree,
 extern gboolean		IDL_tree_property_remove	(IDL_tree tree,
 							 const char *key);
 
+extern void		IDL_tree_walk			(IDL_tree p,
+							 IDL_tree parent,
+							 IDL_tree_func pre_tree_func,
+							 IDL_tree_func post_tree_func,
+							 gpointer user_data);
+
 extern void		IDL_tree_walk_in_order		(IDL_tree p,
 							 IDL_tree_func tree_func,
 							 gpointer user_data);
 
 extern void		IDL_tree_free			(IDL_tree root);
 
+extern void		IDL_tree_to_IDL			(IDL_tree p,
+							 IDL_ns ns,
+							 FILE *output,
+							 unsigned long output_flags);
+
 extern char *		IDL_do_escapes			(const char *s);
 
 extern IDL_tree		IDL_resolve_const_exp		(IDL_tree p,
@@ -758,6 +778,10 @@ extern char *		IDL_ns_ident_to_qstring		(IDL_tree ns_ident,
 							 const char *join,
 							 int scope_levels);
 
+extern int		IDL_ns_scope_levels_from_here	(IDL_ns ns,
+							 IDL_tree ident,
+							 IDL_tree parent);
+
 extern char *		IDL_ns_ident_make_repo_id	(IDL_ns ns,
 							 IDL_tree p,
 							 const char *p_prefix,
