@@ -59,7 +59,7 @@ static gint giop_recv_locate_reply_decode_message(GIOPRecvBuffer *buf);
 static gint giop_recv_locate_request_decode_message(GIOPRecvBuffer *buf);
 static GIOPRecvBuffer *giop_received_list_check_reply(GIOP_unsigned_long request_id);
 
-#ifdef _REENTRANT
+#ifdef NOT_REENTRANT
 extern DEFINE_LOCK(iiop_connection_list);
 #endif
 GList *iiop_connection_list = NULL;
@@ -300,10 +300,10 @@ giop_send_request_buffer_use(GIOPConnection *connection,
   g_return_val_if_fail(object_key_vec, NULL);
   g_return_val_if_fail(operation_vec, NULL);
 
- ORBit_Trace(TraceMod_IIOP, TraceLevel_Debug,
-             "Sending request %s id %d to %s\n",
-	     operation_vec->iov_base + 4,
-	     request_id, object_key_vec->iov_base + 4);
+  ORBit_Trace(TraceMod_IIOP, TraceLevel_Debug,
+	      "Sending request %s id %d to %s\n",
+	      operation_vec->iov_base + 4,
+	      request_id, object_key_vec->iov_base + 4);
 
   send_buffer = giop_send_buffer_use(connection);
 
@@ -490,6 +490,50 @@ gpointer
 giop_send_buffer_append_mem_indirect(GIOPSendBuffer *send_buffer,
 				     gconstpointer mem_region,
 				     gulong mem_region_length)
+{
+  gulong offset = 0, thisblock_size;
+  gpointer blockstart = NULL;
+
+  while(offset < mem_region_length) {
+    thisblock_size = MIN(mem_region_length - offset,
+			 GIOP_INDIRECT_CHUNK_SIZE - send_buffer->indirect_used);
+
+    if(thisblock_size >= sizeof(CORBA_unsigned_long)
+       || (mem_region_length - offset) < sizeof(CORBA_unsigned_long)) {
+      if(!blockstart)
+	blockstart = send_buffer->indirect + send_buffer->indirect_used;
+
+      memcpy(send_buffer->indirect + send_buffer->indirect_used,
+	     mem_region + offset, thisblock_size);
+      giop_message_buffer_append_mem(GIOP_MESSAGE_BUFFER(send_buffer),
+				     send_buffer->indirect +
+				     send_buffer->indirect_used,
+				     thisblock_size);
+      offset += thisblock_size;
+      send_buffer->indirect_used += thisblock_size;
+    }
+
+    if(send_buffer->indirect_used > GIOP_INDIRECT_CHUNK_SIZE
+       || (thisblock_size >= sizeof(CORBA_unsigned_long)
+	   && (mem_region_length - offset) > 0)) {
+	  send_buffer->indirect_used = 0;
+	  send_buffer->indirect = g_chunk_new(gpointer,
+					      send_buffer->indirects);
+#ifdef DEBUG
+	  memset(send_buffer->indirect, '\xFE', GIOP_INDIRECT_CHUNK_SIZE);
+#endif
+
+    }
+  }
+
+  return blockstart;
+}
+
+#ifdef WE_WANT_OLD_DEAD_CRAPPY_BUGGY_CODE
+gpointer
+_giop_send_buffer_append_mem_indirect(GIOPSendBuffer *send_buffer,
+				     gconstpointer mem_region,
+				     gulong mem_region_length)
 {
   gpointer blockstart = NULL;
   gulong offset, new_offset;
@@ -507,18 +551,18 @@ giop_send_buffer_append_mem_indirect(GIOPSendBuffer *send_buffer,
 	  blockstart = send_buffer->indirect + send_buffer->indirect_used;
       }
 	
-	memcpy(send_buffer->indirect + send_buffer->indirect_used,
-	       mem_region + offset, new_offset - offset);
+      memcpy(send_buffer->indirect + send_buffer->indirect_used,
+	     mem_region + offset, new_offset - offset);
 	
-	giop_message_buffer_append_mem(GIOP_MESSAGE_BUFFER(send_buffer),
-				       send_buffer->indirect + send_buffer->indirect_used,
-				       new_offset - offset);
+      giop_message_buffer_append_mem(GIOP_MESSAGE_BUFFER(send_buffer),
+				     send_buffer->indirect + send_buffer->indirect_used,
+				     new_offset - offset);
 
-	send_buffer->indirect_used += new_offset - offset;
+      send_buffer->indirect_used += new_offset - offset;
 
       offset = new_offset;
 
-      if(new_offset >= mem_region_length)
+      if(new_offset >= GIOP_INDIRECT_CHUNK_SIZE)
 	{
 	  send_buffer->indirect_used = 0;
 	  send_buffer->indirect = g_chunk_new(gpointer,
@@ -531,6 +575,7 @@ giop_send_buffer_append_mem_indirect(GIOPSendBuffer *send_buffer,
 
   return blockstart;
 }
+#endif
 
 gpointer
 giop_send_buffer_append_mem_indirect_a(GIOPSendBuffer *send_buffer,
@@ -732,7 +777,6 @@ void
 giop_received_list_push(GIOPRecvBuffer *recv_buffer)
 {
   GET_LOCK(incoming_bufs);
-  g_message("[%d] Adding buffer %p onto received list", getpid(), recv_buffer);
   incoming_bufs = g_list_prepend(incoming_bufs, recv_buffer);
   RELEASE_LOCK(incoming_bufs);
 }
@@ -768,10 +812,11 @@ giop_received_list_check_reply(GIOP_unsigned_long request_id)
 
   for(item = incoming_bufs; item; item = g_list_next(item))
     {
-      retval = item->data;
-      if(GIOP_MESSAGE_BUFFER(retval)->message_header.message_type == GIOP_REPLY
-	 && retval->message.u.reply.request_id == request_id)
+      if(GIOP_MESSAGE_BUFFER(item->data)->message_header.message_type == GIOP_REPLY
+	 && GIOP_RECV_BUFFER(item->data)->message.u.reply.request_id == request_id) {
+	retval = item->data;
 	break;
+      }
     }
 
   if(retval)
@@ -791,6 +836,10 @@ giop_recv_reply_buffer_use_multiple(GArray *request_ids,
   return giop_recv_reply_buffer_use_multiple_2(NULL, request_ids, block_for_reply);
 }
 
+/* here is how it will be:
+   one routine for getting next message with a specified reply ID.
+ */
+
 GIOPRecvBuffer *
 giop_recv_reply_buffer_use_multiple_2(GIOPConnection *request_cnx,
 				      GArray *request_ids,
@@ -798,6 +847,7 @@ giop_recv_reply_buffer_use_multiple_2(GIOPConnection *request_cnx,
 {
   int i;
   GIOPRecvBuffer *retval = NULL;
+  GSList *pushme = NULL;
 
   do {
     /*
@@ -818,13 +868,15 @@ giop_recv_reply_buffer_use_multiple_2(GIOPConnection *request_cnx,
 			(GIOP_unsigned_long *)request_ids->data,
 			request_ids->len))
 	   break;
-	 else
-	  giop_received_list_push(retval);
+	 else {
+	   pushme = g_slist_prepend(pushme, retval); retval = NULL;
+	 }
       } else {
 	if(IIOPIncomingMessageHandler)
 	  IIOPIncomingMessageHandler(retval);
-	else
-	  giop_received_list_push(retval);
+	else {
+	  pushme = g_slist_prepend(pushme, retval); retval = NULL;
+	}
 	retval = NULL;
       }
     } else
@@ -832,7 +884,9 @@ giop_recv_reply_buffer_use_multiple_2(GIOPConnection *request_cnx,
 
   } while(!retval && block_for_reply);
 
-  g_message("[%d] returning buffer %p from recv_reply", getpid(), retval);
+  g_slist_foreach(pushme, (GFunc)giop_received_list_push, NULL);
+  g_slist_free(pushme);
+
   return retval;
 }
 
