@@ -15,6 +15,14 @@
 /* Input: main part of filename to write, namespace & tree from parser
    Output: (in files) header, skeletons, stubs, and common code.
 */
+#ifndef HAVE_INDENT
+#define INDENT "cat"
+#elif HAVE_INDENT_KR
+#define INDENT "indent -npro -kr -i8 -bad -fca -sc -sob"
+#else
+#define INDENT "indent -npro -bad -bap -bc -sob -br -ce -cli2 -npcs -di1 -psl -i3 -lp"
+#endif /* HAVE_INDENT */
+
 void
 orbit_IDL_tree_to_c(const char *base_filename, IDL_ns namespace, IDL_tree tree)
 {
@@ -22,50 +30,65 @@ orbit_IDL_tree_to_c(const char *base_filename, IDL_ns namespace, IDL_tree tree)
   GString *curfile_name = g_string_new(NULL);
   char *hdrname;
 
+  if(debuglevel >= 4)
+    orbit_cbe_print_node(tree, 0);
+  
   base_filename = g_basename(base_filename);
-  orbit_cbe_print_node(tree, 0);
 
-  g_string_sprintf(curfile_name, "%s.h", base_filename);
-  curfile = fopen(curfile_name->str, "w");
+  g_string_sprintf(curfile_name, INDENT " > %s.h", base_filename);
+  curfile = popen(curfile_name->str, "w");
   if(!curfile) {
-    g_warning("Failed to open %s for writing: %s\n",
+    g_warning("Failed to open %s for writing: %s",
 	      curfile_name->str, g_strerror(errno));
     return;
   }
   orbit_cbe_write_header(curfile, namespace, tree);
+  g_string_sprintf(curfile_name, "%s.h", base_filename);
   hdrname = g_strdup(g_basename(curfile_name->str));
   fclose(curfile);
 
-  g_string_sprintf(curfile_name, "%s-common.c", base_filename);
-  curfile = fopen(curfile_name->str, "w");
+  g_string_sprintf(curfile_name, INDENT " > %s-common.c", base_filename);
+  curfile = popen(curfile_name->str, "w");
   if(!curfile) {
-    g_warning("Failed to open %s for writing: %s\n",
+    g_warning("Failed to open %s for writing: %s",
 	      curfile_name->str, g_strerror(errno));
     return;
   }
   orbit_cbe_write_common(curfile, namespace, tree, hdrname);
   fclose(curfile);
 
-  g_string_sprintf(curfile_name, "%s-stubs.c", base_filename);
-  curfile = fopen(curfile_name->str, "w");
+  g_string_sprintf(curfile_name, INDENT " > %s-stubs.c", base_filename);
+  curfile = popen(curfile_name->str, "w");
   if(!curfile) {
-    g_warning("Failed to open %s for writing: %s\n",
+    g_warning("Failed to open %s for writing: %s",
 	      curfile_name->str, g_strerror(errno));
     return;
   }
   orbit_cbe_write_stubs(curfile, namespace, tree, hdrname);
   fclose(curfile);
 
-  g_string_sprintf(curfile_name, "%s-skels.c", base_filename);
-  curfile = fopen(curfile_name->str, "w");
+  g_string_sprintf(curfile_name, INDENT " > %s-skels.c", base_filename);
+  curfile = popen(curfile_name->str, "w");
   if(!curfile) {
-    g_warning("Failed to open %s for writing: %s\n",
+    g_warning("Failed to open %s for writing: %s",
 	      curfile_name->str, g_strerror(errno));
     return;
   }
   orbit_cbe_write_skeletons(curfile, namespace, tree, hdrname);
   fclose(curfile);
 
+  if(enable_skeleton_impl) {
+    g_string_sprintf(curfile_name, INDENT " > %s-impl.c", base_filename);
+    curfile = popen(curfile_name->str, "w");
+    if(!curfile) {
+      g_warning("Failed to open %s for writing: %s",
+		curfile_name->str, g_strerror(errno));
+      return;
+    }
+    orbit_cbe_write_skelimpl(curfile, namespace, tree, hdrname);
+    fclose(curfile);
+  }
+
   g_string_free(curfile_name, TRUE);
   g_free(hdrname);
 }
@@ -88,6 +111,8 @@ orbit_cbe_print_node(IDL_tree node, int indent_level)
     return;
   }
 
+  g_print("[%d] ", IDL_NODE_REFS(node));
+
   switch(IDL_NODE_TYPE(node)) {
 
   case IDLN_NONE:
@@ -104,14 +129,17 @@ orbit_cbe_print_node(IDL_tree node, int indent_level)
 
   case IDLN_GENTREE:
     g_print("GENTREE:\n");
+#if 0
+    /* Changed in libIDL.  But don't need it here anyway. */
     orbit_cbe_print_node(IDL_GENTREE(node).data, indent_level + 2);
     indent(indent_level + 2);
     g_print("children:\n");
     orbit_cbe_print_node(IDL_GENTREE(node).children, indent_level + 4);
+#endif
     break;
 
   case IDLN_INTEGER:
-    g_print("INTEGER: " IDL_SB10_FMT "\n", IDL_INTEGER(node).value);
+    g_print("INTEGER: %" IDL_LL "d\n", IDL_INTEGER(node).value);
     break;
 
   case IDLN_STRING:
@@ -144,8 +172,9 @@ orbit_cbe_print_node(IDL_tree node, int indent_level)
 
   case IDLN_IDENT:
     s = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(node), "_", 0);
-    g_print("IDENT: %s [%d] NSQ: %s\n", IDL_IDENT(node).str,
-	    IDL_IDENT(node)._refs, s);
+    g_print("IDENT: %s NSQ: %s RID: \"%s\"\n",
+	    IDL_IDENT(node).str, s,
+	    IDL_IDENT_REPO_ID(node) ? IDL_IDENT_REPO_ID(node) : "");
     free(s);
     break;
 
@@ -340,31 +369,40 @@ orbit_cbe_print_node(IDL_tree node, int indent_level)
   case IDLN_TYPE_ANY:
     g_print("TYPE any\n");
     break;
+  case IDLN_TYPE_TYPECODE:
+    g_print("TYPE TypeCode\n");
+    break;
   default:
     g_print("unhandled %d\n", IDL_NODE_TYPE(node));
   }
 }
 
-
-void
-orbit_cbe_write_typename(FILE *of, IDL_tree tree)
+char *
+orbit_cbe_get_typename(IDL_tree tree)
 {
+  GString *tmpstr;
+  char *retval = NULL;
+
   if(!tree) {
-    fprintf(of, "void");
-    return;
+    return g_strdup("void");
   }
 
+  tmpstr = g_string_new(NULL);
+
   switch(IDL_NODE_TYPE(tree)) {
+  case IDLN_TYPE_ANY:
+    retval = "CORBA_any";
+    break;
   case IDLN_TYPE_FLOAT:
     switch(IDL_TYPE_FLOAT(tree).f_type) {
     case IDL_FLOAT_TYPE_FLOAT:
-      fprintf(of, "CORBA_float");
+      retval = "CORBA_float";
       break;
     case IDL_FLOAT_TYPE_DOUBLE:
-      fprintf(of, "CORBA_double");
+      retval = "CORBA_double";
       break;
     case IDL_FLOAT_TYPE_LONGDOUBLE:
-      fprintf(of, "CORBA_long_double");
+      retval = "CORBA_long_double";
       break;
     }
     break;
@@ -374,64 +412,111 @@ orbit_cbe_write_typename(FILE *of, IDL_tree tree)
 	    IDL_TYPE_FIXED(tree).positive_int_const,
 	    IDL_TYPE_FIXED(tree).integer_lit);
 #else
-    fprintf(of, "CORBA_fixed_d_s");
+    retval = "CORBA_fixed_d_s";
 #endif
     break;
   case IDLN_TYPE_INTEGER:
-    fprintf(of, "CORBA_");
+    g_string_assign(tmpstr, "CORBA_");
     if(!IDL_TYPE_INTEGER(tree).f_signed)
-      fprintf(of, "unsigned_");
+      g_string_append(tmpstr, "unsigned_");
+
     switch(IDL_TYPE_INTEGER(tree).f_type) {
     case IDL_INTEGER_TYPE_SHORT:
-      fprintf(of, "short");
+      g_string_append(tmpstr, "short");
       break;
     case IDL_INTEGER_TYPE_LONGLONG:
-      fprintf(of, "long_");
+      g_string_append(tmpstr, "long_");
     case IDL_INTEGER_TYPE_LONG:
-      fprintf(of, "long");
+      g_string_append(tmpstr, "long");
       break;
     }
     break;
   case IDLN_TYPE_STRING:
-    fprintf(of, "CORBA_string");
+    retval = "CORBA_string";
     break;
   case IDLN_TYPE_OCTET:
-    fprintf(of, "CORBA_octet");
+    retval = "CORBA_octet";
     break;
   case IDLN_TYPE_WIDE_STRING:
-    fprintf(of, "CORBA_wstring");
+    retval = "CORBA_wstring";
     break;
   case IDLN_TYPE_CHAR:
-    fprintf(of, "CORBA_char");
+    retval = "CORBA_char";
     break;
   case IDLN_TYPE_WIDE_CHAR:
-    fprintf(of, "CORBA_wchar");
+    retval = "CORBA_wchar";
+    break;
+  case IDLN_TYPE_BOOLEAN:
+    retval = "CORBA_boolean";
     break;
   case IDLN_TYPE_STRUCT:
-    orbit_cbe_write_typename(of, IDL_TYPE_STRUCT(tree).ident);
+    retval = orbit_cbe_get_typename(IDL_TYPE_STRUCT(tree).ident);
+    g_string_assign(tmpstr, retval); free(retval); retval = NULL;
     break;
   case IDLN_EXCEPT_DCL:
-    orbit_cbe_write_typename(of, IDL_EXCEPT_DCL(tree).ident);
+    retval = orbit_cbe_get_typename(IDL_EXCEPT_DCL(tree).ident);
+    g_string_assign(tmpstr, retval); free(retval); retval = NULL;
     break;
-  case IDLN_TYPE_ARRAY:    
-    orbit_cbe_write_typename(of, IDL_TYPE_ARRAY(tree).ident);
+  case IDLN_TYPE_ARRAY:
+    retval = orbit_cbe_get_typename(IDL_TYPE_ARRAY(tree).ident);
+    g_string_assign(tmpstr, retval); free(retval); retval = NULL;
     break;
   case IDLN_TYPE_UNION:
-    orbit_cbe_write_typename(of, IDL_TYPE_UNION(tree).ident);
+    retval = orbit_cbe_get_typename(IDL_TYPE_UNION(tree).ident);
+    g_string_assign(tmpstr, retval); free(retval); retval = NULL;
     break;
   case IDLN_TYPE_ENUM:
-    orbit_cbe_write_typename(of, IDL_TYPE_ENUM(tree).ident);
+    retval = orbit_cbe_get_typename(IDL_TYPE_ENUM(tree).ident);
+    g_string_assign(tmpstr, retval); free(retval); retval = NULL;
     break;
   case IDLN_IDENT:
     {
       char *id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(tree), "_", 0);
-      fprintf(of, "%s", id);
+      g_string_assign(tmpstr, id);
       free(id);
     }
     break;
+  case IDLN_PARAM_DCL:
+    return orbit_cbe_get_typename(IDL_PARAM_DCL(tree).param_type_spec);
+    break;
+  case IDLN_TYPE_SEQUENCE:
+    return orbit_cbe_get_typename(IDL_get_parent_node(tree, IDLN_TYPE_DCL, NULL));
+    break;
+  case IDLN_INTERFACE:
+    retval = orbit_cbe_get_typename(IDL_INTERFACE(tree).ident);
+    g_string_assign(tmpstr, retval); free(retval); retval = NULL;
+    break;
+  case IDLN_NATIVE:
+    retval = "gpointer";
+    break;
+  case IDLN_TYPE_OBJECT:
+    retval = "CORBA_Object";
+    break;
+  case IDLN_TYPE_TYPECODE:
+    retval = "CORBA_TypeCode";
+    break;
   default:
-    g_error("We were asked to print a typespec %s\n", IDL_tree_type_names[tree->_type]);
+    g_error("We were asked to get a typename for a %s",
+	    IDL_tree_type_names[IDL_NODE_TYPE(tree)]);
   }
+
+  if(retval)
+    return g_strdup(retval);
+  else {
+    retval = g_strdup(tmpstr->str);
+    g_string_free(tmpstr, TRUE);
+    return retval;
+  }
+}
+
+void
+orbit_cbe_write_typename(FILE *of, IDL_tree tree)
+{
+  char *t = orbit_cbe_get_typename(tree);
+
+  fprintf(of, "%s", t);
+
+  free(t);
 }
 
 void
@@ -456,6 +541,9 @@ orbit_cbe_write_typespec(FILE *of, IDL_tree tree)
       break;
     };
     break;
+  case IDLN_TYPE_BOOLEAN:
+    fprintf(of, "CORBA_boolean");
+    break;
   case IDLN_TYPE_FIXED:
 #if 0
     fprintf(of, "CORBA_fixed_%d_%d",
@@ -514,8 +602,25 @@ orbit_cbe_write_typespec(FILE *of, IDL_tree tree)
       free(id);
     }
     break;
+  case IDLN_TYPE_OBJECT:
+    fprintf(of, "CORBA_Object");
+    break;
+  case IDLN_TYPE_SEQUENCE:
+    fprintf(of, "struct { CORBA_unsigned_long _maximum, _length;\n");
+    orbit_cbe_write_typespec(of, IDL_TYPE_SEQUENCE(tree).simple_type_spec);
+    fprintf(of, " *_buffer;\nCORBA_boolean _release; } \n");
+    break;
+  case IDLN_TYPE_ANY:
+    fprintf(of, "CORBA_any");
+    break;
+  case IDLN_NATIVE:
+    fprintf(of, "gpointer");
+    break;
+  case IDLN_TYPE_TYPECODE:
+    fprintf(of, "CORBA_TypeCode");
+    break;
   default:
-    g_error("We were asked to print a typespec %s\n", IDL_tree_type_names[tree->_type]);
+    g_error("We were asked to print a typespec %s", IDL_tree_type_names[tree->_type]);
   }
 }
 
@@ -535,7 +640,7 @@ orbit_cbe_write_const(FILE *of, IDL_tree tree)
     fprintf(of, "%f", IDL_FLOAT(tree).value);
     break;
   case IDLN_INTEGER:
-    fprintf(of, "%lld", IDL_INTEGER(tree).value);
+    fprintf(of, "%" IDL_LL "d", IDL_INTEGER(tree).value);
     break;
   case IDLN_STRING:
     fprintf(of, "\"%s\"", IDL_STRING(tree).value);
@@ -594,11 +699,59 @@ orbit_cbe_write_const(FILE *of, IDL_tree tree)
     fprintf(of, "%s", opc);
     orbit_cbe_write_const(of, IDL_UNARYOP(tree).operand);
     break;
+  case IDLN_IDENT:
+    {
+      /* XXX fixme, brokenness */
+      char *id;
+      id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(tree), "_", 0);
+      fprintf(of, "%s", id);
+      free(id);
+    }
+    break;
   default:
-    g_error("We were asked to print a constant for %s\n", IDL_tree_type_names[tree->_type]);
+    g_error("We were asked to print a constant for %s", IDL_tree_type_names[tree->_type]);
   }
 }
 
+static const int * const
+orbit_cbe_get_typeoffsets_table (void)
+{
+  static int typeoffsets[IDLN_LAST];
+  static gboolean initialized = FALSE;
+  
+  if (!initialized) {
+    int i;
+
+    for (i = 0; i < IDLN_LAST; ++i)
+      typeoffsets[i] = -1;
+
+    typeoffsets[IDLN_FORWARD_DCL] = 8; /* (same as objref) */
+    typeoffsets[IDLN_TYPE_INTEGER] = 0;
+    typeoffsets[IDLN_TYPE_FLOAT] = 0;
+    typeoffsets[IDLN_TYPE_FIXED] = 3;
+    typeoffsets[IDLN_TYPE_CHAR] = 5;
+    typeoffsets[IDLN_TYPE_WIDE_CHAR] = 6;
+    typeoffsets[IDLN_TYPE_STRING] = 12;
+    typeoffsets[IDLN_TYPE_WIDE_STRING] = 13;
+    typeoffsets[IDLN_TYPE_BOOLEAN] = 4;
+    typeoffsets[IDLN_TYPE_OCTET] = 7;
+    typeoffsets[IDLN_TYPE_ANY] = 16;
+    typeoffsets[IDLN_TYPE_OBJECT] = 9;
+    typeoffsets[IDLN_TYPE_TYPECODE] = 9;
+    typeoffsets[IDLN_TYPE_ENUM] = 8;
+    typeoffsets[IDLN_TYPE_SEQUENCE] = 14;
+    typeoffsets[IDLN_TYPE_ARRAY] = 15;
+    typeoffsets[IDLN_TYPE_STRUCT] = 10;
+    typeoffsets[IDLN_TYPE_UNION] = 11;
+    typeoffsets[IDLN_NATIVE] = 15; /* no pointers ever, same as fixed array */
+    typeoffsets[IDLN_INTERFACE] = 9; /* (same as objref) */
+    
+    initialized = TRUE;
+  }
+  
+  return typeoffsets;
+}
+
 /*******
 
 	This is a rather hairy function. Its purpose is to output the
@@ -613,50 +766,7 @@ orbit_cbe_write_const(FILE *of, IDL_tree tree)
 gint
 orbit_cbe_param_numptrs(IDL_tree param, IDL_ParamRole role)
 {
-  const int typeoffsets[] = {
-    -1, /* IDLN_NONE, */
-    -1, /* IDLN_ANY, */
-    -1, /* IDLN_LIST, */
-    -1, /* IDLN_GENTREE, */
-    -1, /* IDLN_INTEGER, */
-    -1, /* IDLN_STRING, */
-    -1, /* IDLN_WIDE_STRING, */
-    -1, /* IDLN_CHAR, */
-    -1, /* IDLN_WIDE_CHAR, */
-    -1, /* IDLN_FIXED, */
-    -1, /* IDLN_FLOAT, */
-    -1, /* IDLN_BOOLEAN, */
-    -1, /* IDLN_IDENT, */
-    -1, /* IDLN_TYPE_DCL, */
-    -1, /* IDLN_CONST_DCL, */
-    -1, /* IDLN_EXCEPT_DCL, */
-    -1, /* IDLN_ATTR_DCL, */
-    -1, /* IDLN_OP_DCL, */
-    -1, /* IDLN_PARAM_DCL, */
-    -1, /* IDLN_FORWARD_DCL, */
-    0, /* IDLN_TYPE_INTEGER, */
-    0, /* IDLN_TYPE_FLOAT, */
-    3, /* IDLN_TYPE_FIXED, */
-    5, /* IDLN_TYPE_CHAR, */
-    6, /* IDLN_TYPE_WIDE_CHAR, */
-    12, /* IDLN_TYPE_STRING, */
-    13, /*     IDLN_TYPE_WIDE_STRING, */
-    4, /*     IDLN_TYPE_BOOLEAN, */
-    7, /*     IDLN_TYPE_OCTET, */
-    16, /*     IDLN_TYPE_ANY, */
-    9, /*     IDLN_TYPE_OBJECT, */
-    8, /*     IDLN_TYPE_ENUM, */
-    14, /*     IDLN_TYPE_SEQUENCE, */
-    15, /*     IDLN_TYPE_ARRAY, */
-    10, /*     IDLN_TYPE_STRUCT, */
-    11, /*     IDLN_TYPE_UNION, */
-    -1, /* IDLN_MEMBER, */
-    -1, /* IDLN_CASE_STMT, */
-    -1, /* IDLN_INTERFACE, */
-    -1, /* IDLN_MODULE, */
-    -1, /* IDLN_BINOP, */
-    -1, /* IDLN_UNARYOP */
-  };
+  const int * const typeoffsets = orbit_cbe_get_typeoffsets_table ();
   const int nptrrefs_required[][4] = {
     {0,1,1,0} /* float */,
     {0,1,1,0} /* double */,
@@ -673,7 +783,7 @@ orbit_cbe_param_numptrs(IDL_tree param, IDL_ParamRole role)
     {0,1,1,0} /* string */,
     {0,1,1,0} /* wstring */,
     {1,1,2,1} /* sequence */,
-    {0,0,0,1} /* fixed array */,
+    {0,0,0,0} /* fixed array */,
     {1,1,2,1} /* any 16 */
   };
   int retval = 0;
@@ -686,13 +796,6 @@ orbit_cbe_param_numptrs(IDL_tree param, IDL_ParamRole role)
 
   g_assert(param);
 
-  switch(role) {
-  case DATA_IN: role = 0; break;
-  case DATA_INOUT: role = 1; break;
-  case DATA_OUT: role = 2; break;
-  case DATA_RETURN: role = 3; break;
-  }
-
   switch(IDL_NODE_TYPE(param))
     {
     case IDLN_TYPE_STRUCT:
@@ -700,41 +803,49 @@ orbit_cbe_param_numptrs(IDL_tree param, IDL_ParamRole role)
       if(((role == DATA_RETURN) || (role == DATA_OUT))
 	 && !cbe_type_is_fixed_length(param))
 	retval++;
-      break;
+
       break;
     case IDLN_TYPE_ARRAY:
       if(!cbe_type_is_fixed_length(param) && role == DATA_OUT)
-	retval += 2;
+	retval++;
       break;
     default:
     }
 
   g_assert(typeoffsets[IDL_NODE_TYPE(param)] >= 0);
 
-  return retval+nptrrefs_required[typeoffsets[IDL_NODE_TYPE(param)]][role];
+  switch(role) {
+  case DATA_IN: role = 0; break;
+  case DATA_INOUT: role = 1; break;
+  case DATA_OUT: role = 2; break;
+  case DATA_RETURN: role = 3; break;
+  }
+
+  retval+=nptrrefs_required[typeoffsets[IDL_NODE_TYPE(param)]][role];
+
+  return retval;
 }
 
 void
 orbit_cbe_param_printptrs(FILE *of, IDL_tree param, IDL_ParamRole role)
 {
-  int i;
+  int i, n;
 
   if(param == NULL)
     return;
 
   param = cbe_get_typespec(param);
+  n = orbit_cbe_param_numptrs(param, role);
 
   if(IDL_NODE_TYPE(param) == IDLN_TYPE_ARRAY
      && ((role == DATA_OUT) || (role == DATA_RETURN))) {
     if(role == DATA_RETURN)
-      fprintf(of, "_slice");
-    if(!cbe_type_is_fixed_length(param)) {
-      if(role == DATA_OUT)
-	fprintf(of, "_slice");
-    }
+      fprintf(of, "_slice*");
+    else if(!cbe_type_is_fixed_length(param)) /* && role == DATA_OUT */
+      fprintf(of, "_slice*");
   }
 
-  for(i = 0; i < orbit_cbe_param_numptrs(param, role); i++)
+  for(i = 0; i < n; i++)
     fprintf(of, "*");
 }
 
@@ -745,27 +856,36 @@ cbe_type_is_fixed_length(IDL_tree ts)
   IDL_tree curitem;
 
   ts = cbe_get_typespec(ts);
-
   switch(IDL_NODE_TYPE(ts)) {
   case IDLN_TYPE_FLOAT:
   case IDLN_TYPE_INTEGER:
   case IDLN_TYPE_ENUM:
   case IDLN_TYPE_CHAR:
   case IDLN_TYPE_WIDE_CHAR:
+  case IDLN_TYPE_OCTET:
+  case IDLN_TYPE_BOOLEAN:
     return TRUE;
     break;
   case IDLN_TYPE_SEQUENCE:
   case IDLN_TYPE_STRING:
   case IDLN_TYPE_WIDE_STRING:
+  case IDLN_TYPE_OBJECT:
+  case IDLN_FORWARD_DCL:
+  case IDLN_INTERFACE:
+  case IDLN_TYPE_ANY:
+  case IDLN_NATIVE:
+  case IDLN_TYPE_TYPECODE:
     return FALSE;
     break;
   case IDLN_TYPE_UNION:
+#if 1
     for(curitem = IDL_TYPE_UNION(ts).switch_body; curitem;
 	curitem = IDL_LIST(curitem).next) {
       is_fixed &= cbe_type_is_fixed_length(IDL_LIST(IDL_CASE_STMT(IDL_LIST(curitem).data).element_spec).data);
     }
     return is_fixed;
     break;
+#endif
   case IDLN_EXCEPT_DCL:
   case IDLN_TYPE_STRUCT:
     for(curitem = IDL_TYPE_STRUCT(ts).member_list; curitem;
@@ -774,8 +894,11 @@ cbe_type_is_fixed_length(IDL_tree ts)
     }
     return is_fixed;
     break;
+  case IDLN_TYPE_ARRAY:
+    return cbe_type_is_fixed_length(IDL_TYPE_DCL(IDL_get_parent_node(ts, IDLN_TYPE_DCL, NULL)).type_spec);
+    break;
   default:
-    g_warning("I'm not sure if type %s is fixed-length\n", IDL_tree_type_names[IDL_NODE_TYPE(ts)]);
+    g_warning("I'm not sure if type %s is fixed-length", IDL_tree_type_names[IDL_NODE_TYPE(ts)]);
     return FALSE;
   }
 }
@@ -785,6 +908,7 @@ cbe_get_typespec(IDL_tree node)
 {
   if(node == NULL)
     return NULL;
+
   switch(IDL_NODE_TYPE(node)) {
   case IDLN_TYPE_INTEGER:
   case IDLN_TYPE_FLOAT:
@@ -803,6 +927,9 @@ cbe_get_typespec(IDL_tree node)
   case IDLN_TYPE_STRUCT:
   case IDLN_TYPE_UNION:
   case IDLN_EXCEPT_DCL:
+  case IDLN_INTERFACE:
+  case IDLN_NATIVE:
+  case IDLN_TYPE_TYPECODE:
     return node;
     break;
   case IDLN_TYPE_DCL:
@@ -814,19 +941,86 @@ cbe_get_typespec(IDL_tree node)
   case IDLN_MEMBER:
     return cbe_get_typespec(IDL_MEMBER(node).type_spec);
     break;
+  case IDLN_FORWARD_DCL:
+    return cbe_get_typespec(IDL_FORWARD_DCL(node).ident);
+    break;
   case IDLN_LIST:
   case IDLN_IDENT:
     return cbe_get_typespec(IDL_get_parent_node(node, IDLN_ANY, NULL));
     break;
   default:
-    g_warning("Unhandled node type %s!\n", IDL_tree_type_names[IDL_NODE_TYPE(node)]);
+    g_warning("Unhandled node type %s!", IDL_tree_type_names[IDL_NODE_TYPE(node)]);
     return NULL;
   }
 }
 
 int
-cbe_get_type_alignment(IDL_tree typespec)
+cbe_get_type_head_alignment(IDL_tree typespec)
 {
+  switch(IDL_NODE_TYPE(typespec)) {
+  case IDLN_TYPE_INTEGER:
+    switch(IDL_TYPE_INTEGER(typespec).f_type) {
+    case IDL_INTEGER_TYPE_SHORT: return sizeof(CORBA_short);
+    case IDL_INTEGER_TYPE_LONG: return sizeof(CORBA_long);
+    case IDL_INTEGER_TYPE_LONGLONG: return sizeof(CORBA_long_long);
+    }
+    break;
+  case IDLN_TYPE_FLOAT:
+    switch(IDL_TYPE_FLOAT(typespec).f_type) {
+    case IDL_FLOAT_TYPE_FLOAT: return sizeof(CORBA_float);
+    case IDL_FLOAT_TYPE_DOUBLE: return sizeof(CORBA_double);
+    case IDL_FLOAT_TYPE_LONGDOUBLE: return sizeof(CORBA_long_double);
+    }
+    break;
+  case IDLN_TYPE_FIXED:
+    return sizeof(CORBA_short);
+    break;
+  case IDLN_TYPE_CHAR:
+    return sizeof(CORBA_char);
+    break;
+  case IDLN_TYPE_WIDE_CHAR:
+    return sizeof(CORBA_wchar);
+    break;
+  case IDLN_TYPE_STRING:
+  case IDLN_INTERFACE:
+  case IDLN_TYPE_OBJECT:
+  case IDLN_FORWARD_DCL:
+  case IDLN_TYPE_TYPECODE:
+    return sizeof(CORBA_unsigned_long);
+    break;
+  case IDLN_TYPE_WIDE_STRING:
+  case IDLN_TYPE_BOOLEAN:
+  case IDLN_TYPE_OCTET:
+  case IDLN_NATIVE:
+    return 1;
+    break;
+  case IDLN_TYPE_ANY:
+  case IDLN_TYPE_ENUM:
+  case IDLN_TYPE_SEQUENCE:
+  case IDLN_EXCEPT_DCL:
+    return sizeof(CORBA_unsigned_long);
+    break;
+  case IDLN_TYPE_ARRAY:
+    return cbe_get_type_head_alignment(cbe_get_typespec(IDL_TYPE_ARRAY(typespec).ident));
+    break;
+  case IDLN_TYPE_STRUCT:
+    return cbe_get_type_head_alignment(cbe_get_typespec(IDL_LIST(IDL_TYPE_STRUCT(typespec).member_list).data));
+    break;
+  case IDLN_TYPE_UNION:
+    return cbe_get_type_head_alignment(cbe_get_typespec(IDL_TYPE_UNION(typespec).switch_type_spec));
+    break;
+  default:
+    return cbe_get_type_head_alignment(cbe_get_typespec(typespec));
+  }
+  return 0;
+}
+
+int
+cbe_get_type_tail_alignment(IDL_tree typespec)
+{
+  int minalign;
+  IDL_tree curitem;
+
   switch(IDL_NODE_TYPE(typespec)) {
   case IDLN_TYPE_INTEGER:
     switch(IDL_TYPE_INTEGER(typespec).f_type) {
@@ -855,27 +1049,43 @@ cbe_get_type_alignment(IDL_tree typespec)
   case IDLN_TYPE_WIDE_STRING:
   case IDLN_TYPE_BOOLEAN:
   case IDLN_TYPE_OCTET:
+  case IDLN_INTERFACE:
+  case IDLN_TYPE_OBJECT:
+  case IDLN_FORWARD_DCL:
+  case IDLN_TYPE_ANY:
+  case IDLN_NATIVE:
+  case IDLN_TYPE_TYPECODE:
     return 1;
     break;
-  case IDLN_TYPE_OBJECT:
-    break;
-  case IDLN_TYPE_ANY:
   case IDLN_TYPE_ENUM:
-  case IDLN_TYPE_SEQUENCE:
-  case IDLN_EXCEPT_DCL:
     return sizeof(CORBA_unsigned_long);
     break;
-  case IDLN_TYPE_ARRAY:
-    return cbe_get_type_alignment(cbe_get_typespec(IDL_TYPE_ARRAY(typespec).ident));
+  case IDLN_TYPE_SEQUENCE:
+    return cbe_get_type_tail_alignment(cbe_get_typespec(IDL_TYPE_SEQUENCE(typespec).simple_type_spec));
     break;
+  case IDLN_TYPE_ARRAY:
+    return cbe_get_type_tail_alignment(cbe_get_typespec(IDL_TYPE_ARRAY(typespec).ident));
+    break;
+  case IDLN_EXCEPT_DCL:
   case IDLN_TYPE_STRUCT:
-    return cbe_get_type_alignment(cbe_get_typespec(IDL_LIST(IDL_TYPE_STRUCT(typespec).member_list).data));
+    curitem = IDL_list_nth(IDL_TYPE_STRUCT(typespec).member_list,
+			   IDL_list_length(IDL_TYPE_STRUCT(typespec).member_list) - 1);
+    return cbe_get_type_tail_alignment(cbe_get_typespec(IDL_LIST(curitem).data));
     break;
   case IDLN_TYPE_UNION:
-    return cbe_get_type_alignment(cbe_get_typespec(IDL_TYPE_UNION(typespec).switch_type_spec));
+    for(minalign = 8 /* FIXME */, curitem = IDL_TYPE_UNION(typespec).switch_body;
+	curitem && minalign > 1;
+	curitem = IDL_LIST(curitem).next) {
+      IDL_tree uts;
+
+      uts = IDL_CASE_STMT(IDL_LIST(curitem).data).element_spec;
+
+      minalign = MIN(cbe_get_type_tail_alignment(cbe_get_typespec(uts)), minalign);
+    }
+    return minalign;
     break;
   default:
-    g_warning("Unhandled node type %s!\n", IDL_tree_type_names[IDL_NODE_TYPE(typespec)]);
+    return cbe_get_type_tail_alignment(cbe_get_typespec(typespec));
   }
   return 0;
 }
@@ -920,3 +1130,47 @@ IDL_tree_traverse_parents(IDL_tree p,
 
 	g_hash_table_destroy(visited_nodes);
 }
+
+IDL_ParamRole
+cbe_attr_to_paramrole(enum IDL_param_attr attr)
+{
+  switch(attr) {
+  case IDL_PARAM_IN:
+    return DATA_IN;
+  case IDL_PARAM_OUT:
+    return DATA_OUT;
+  case IDL_PARAM_INOUT:
+    return DATA_INOUT;
+  default:
+    g_warning("Unknown IDL_param_attr %d", attr);
+    return -1;
+  }
+}
+
+gboolean
+cbe_type_is_endian_sensitive(IDL_tree typespec)
+{
+  IDL_tree curitem;
+  gboolean retval;
+
+  switch(IDL_NODE_TYPE(typespec)) {
+  case IDLN_TYPE_CHAR:
+  case IDLN_TYPE_BOOLEAN:
+  case IDLN_TYPE_OCTET:
+    return FALSE;
+  case IDLN_TYPE_ARRAY:
+    curitem = IDL_get_parent_node(typespec, IDLN_ANY, NULL);
+    curitem = IDL_get_parent_node(curitem, IDLN_ANY, NULL);
+    return cbe_type_is_endian_sensitive(IDL_MEMBER(curitem).type_spec);
+  case IDLN_TYPE_STRUCT:
+    retval = TRUE;
+    for(curitem = IDL_TYPE_STRUCT(typespec).member_list; curitem;
+	curitem = IDL_LIST(curitem).next) {
+      retval = retval && cbe_type_is_endian_sensitive(IDL_LIST(curitem).data);
+    }
+    return retval;
+    break;
+  default:
+    return TRUE;
+  }
+}
