@@ -8,10 +8,13 @@ typedef struct {
   FILE *of;
   IDL_ns ns;
   IDL_tree tree;
+  char *free_internal;
 } CBESkelInfo;
 
 static void orbit_cbe_skel_process_piece(CBESkelInfo *ski);
 static void cbe_skel_do_attr_dcl(CBESkelInfo *ski);
+static void cbe_skel_do_inherited_attr_dcl(CBESkelInfo *ski,
+					   IDL_tree current_interface);
 static void cbe_skel_do_binop(CBESkelInfo *ski);
 static void cbe_skel_do_boolean(CBESkelInfo *ski);
 static void cbe_skel_do_case_stmt(CBESkelInfo *ski);
@@ -30,6 +33,11 @@ static void cbe_skel_do_member(CBESkelInfo *ski);
 static void cbe_skel_do_module(CBESkelInfo *ski);
 static void cbe_skel_do_none(CBESkelInfo *ski);
 static void cbe_skel_do_op_dcl(CBESkelInfo *ski);
+static void cbe_skel_do_inherited_op_dcl(CBESkelInfo *ski,
+					 IDL_tree current_interface);
+static void cbe_skel_op_params_marshal(CBESkelInfo *ski);
+static void cbe_skel_op_params_demarshal(CBESkelInfo *ski);
+static void cbe_skel_op_params_free(CBESkelInfo *ski);
 static void cbe_skel_do_param_dcl(CBESkelInfo *ski);
 static void cbe_skel_do_string(CBESkelInfo *ski);
 static void cbe_skel_do_type_any(CBESkelInfo *ski);
@@ -57,11 +65,14 @@ void
 orbit_cbe_write_skeletons(FILE *outfile, IDL_ns ns, IDL_tree tree,
 			  const char *header_filename)
 {
-  CBESkelInfo ski = {outfile, ns, tree};
+  CBESkelInfo ski = {outfile, ns, tree, "CORBA_TRUE"};
 
+  fprintf(outfile, "/*\n"
+  		   " * This file was generated by orbit-idl - DO NOT EDIT!\n"
+		   " */\n\n");
   fprintf(outfile, "#include <string.h>\n");
   fprintf(outfile, "#include \"%s\"\n", header_filename);
-  fprintf(outfile, "#define GET_ATOM(x) ({ GIOP_RECV_BUFFER(_ORBIT_recv_buffer)->decoder(&x, (GIOP_RECV_BUFFER(_ORBIT_recv_buffer)->cur), sizeof(x)); GIOP_RECV_BUFFER(_ORBIT_recv_buffer)->cur += sizeof(x); })\n");
+  fprintf(outfile, "#define GET_ATOM(x) ({ GIOP_RECV_BUFFER(_ORBIT_recv_buffer)->decoder(&x, (GIOP_RECV_BUFFER(_ORBIT_recv_buffer)->cur), sizeof(x)); GIOP_RECV_BUFFER(_ORBIT_recv_buffer)->cur += sizeof(x); })\n\n");
 
   orbit_cbe_skel_process_piece(&ski);
 }
@@ -194,13 +205,18 @@ orbit_cbe_skel_process_piece(CBESkelInfo *ski)
     cbe_skel_do_wide_string(ski);
     break;
   case IDLN_ANY:
-    g_error("IDLN_ANY not handled\n");
+    g_error("IDLN_ANY not handled");
     break;
+  case IDLN_NATIVE:
+    break;
+  default:
+    g_warning("Unhandled node type %s",
+	      IDL_tree_type_names[IDL_NODE_TYPE(ski->tree)]);
   }
 }
 
 static void
-cbe_skel_do_attr_dcl(CBESkelInfo *ski)
+cbe_skel_do_attr_dcl_internal(CBESkelInfo *ski, IDL_tree current_interface, gboolean inherited)
 {
   IDL_tree curop, curitem;
   GString *attrname = g_string_new(NULL);
@@ -248,7 +264,10 @@ cbe_skel_do_attr_dcl(CBESkelInfo *ski)
 					      IDL_ident_new(strdup("value"))));
 	    }
 	    
-	    orbit_cbe_skel_process_piece(&subski);
+	    if(inherited==TRUE)
+	      cbe_skel_do_inherited_op_dcl(&subski, current_interface);
+	    else
+	      orbit_cbe_skel_process_piece(&subski);
 
 	    /* Restore the fake link to the original in the namespace */
 	    IDL_GENTREE(IDL_IDENT_TO_NS(IDL_LIST(curitem).data)).data = ns_data_save;
@@ -270,78 +289,116 @@ cbe_skel_do_attr_dcl(CBESkelInfo *ski)
   g_string_free(attrname, TRUE);
 }
 
+static void
+cbe_skel_do_attr_dcl(CBESkelInfo *ski)
+{
+  cbe_skel_do_attr_dcl_internal(ski, NULL, FALSE);
+}
+
+static void
+cbe_skel_do_inherited_attr_dcl(CBESkelInfo *ski, IDL_tree current_interface)
+{
+  cbe_skel_do_attr_dcl_internal(ski, current_interface, TRUE);
+}
+
 static void
 cbe_skel_do_binop(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_boolean(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_case_stmt(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_char(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_const_dcl(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_except_dcl(CBESkelInfo *ski)
 {
-  /* Nothing to do in skels */
+  /* Generate a marshaller routine */
+
+  CBEMarshalInfo mi;
+  char *id;
+  GString *tmpstr = g_string_new(NULL);
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_EXCEPT_DCL(ski->tree).ident),
+			       "_", 0);
+
+  fprintf(ski->of, "static void %s_marshal(GIOPSendBuffer *_ORBIT_send_buffer, CORBA_Environment *ev)\n", id);
+
+  fprintf(ski->of, "{\n");
+
+  mi.of = ski->of;
+  mi.previous_param = NULL;
+  mi.param = mi.typespec = ski->tree;
+  mi.require_indirect = FALSE;
+  g_string_sprintf(tmpstr, "(*((%s *)ev->_params))", id);
+  mi.param_name = tmpstr->str;
+
+  cbe_output_marshaller(&mi);
+
+  fprintf(ski->of, "}\n");
+
+  free(id);
+  g_string_free(tmpstr, TRUE);
+
+
   cbe_skel_do_type_struct(ski);
-  g_warning("Finish up exception handling in skeletons\n");
 }
 
 static void
 cbe_skel_do_fixed(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  /* g_assert(!"Not yet implemented"); */
 }
 
 static void
 cbe_skel_do_float(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_forward_dcl(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_gentree(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_ident(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_integer(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 void
@@ -373,76 +430,191 @@ cbe_skel_print_skelptr(FILE *of, IDL_tree tree)
   free(id);
 }
 
+static void
+cbe_skel_interface_print_relayer_internal(CBESkelInfo *ski, IDL_tree curitem, char *iface_id, char *cur_id)
+{
+  IDL_tree curdcl;
+
+  switch(IDL_NODE_TYPE(IDL_LIST(curitem).data)) {
+  case IDLN_OP_DCL:
+    fprintf(ski->of, "  if(!strcmp(_ORBIT_recv_buffer->message.u.request.operation, \"%s\")) {\n",
+      IDL_IDENT(IDL_OP_DCL(IDL_LIST(curitem).data).ident).str);
+    fprintf(ski->of, "    *epv = servant->vepv->%s_epv->%s;\n", iface_id,
+      IDL_IDENT(IDL_OP_DCL(IDL_LIST(curitem).data).ident).str);
+    fprintf(ski->of,
+      "    return (gpointer)skel_%s_%s;\n",
+      cur_id, IDL_IDENT(IDL_OP_DCL(IDL_LIST(curitem).data).ident).str);
+    fprintf(ski->of, "  } else");
+    break;
+
+  case IDLN_ATTR_DCL:
+    for(curdcl = IDL_ATTR_DCL(IDL_LIST(curitem).data).simple_declarations;
+      curdcl; curdcl = IDL_LIST(curdcl).next) {
+
+      fprintf(ski->of, "  if(!strcmp(_ORBIT_recv_buffer->message.u.request.operation, \"_get_%s\")) {\n",
+	IDL_IDENT(IDL_LIST(curdcl).data).str);
+      fprintf(ski->of, "    *epv = servant->vepv->%s_epv->_get_%s;\n", iface_id,
+	IDL_IDENT(IDL_LIST(curdcl).data).str);
+      fprintf(ski->of,
+	"    return (gpointer)skel_%s__get_%s;\n",
+	cur_id,
+	IDL_IDENT(IDL_LIST(curdcl).data).str);
+      fprintf(ski->of, "  } else");
+
+      if(!IDL_ATTR_DCL(IDL_LIST(curitem).data).f_readonly) {
+        fprintf(ski->of,
+	  "  if(!strcmp(_ORBIT_recv_buffer->message.u.request.operation, \"_set_%s\")) {\n",
+	  IDL_IDENT(IDL_LIST(curdcl).data).str);
+        fprintf(ski->of, "    *epv = servant->vepv->%s_epv->_set_%s;\n", iface_id,
+	  IDL_IDENT(IDL_LIST(curdcl).data).str);
+        fprintf(ski->of,
+	  "    return (gpointer)skel_%s__set_%s;\n",
+	  cur_id, IDL_IDENT(IDL_LIST(curdcl).data).str);
+        fprintf(ski->of, "  } else");
+      }
+    }
+    break;
+  default:
+  }
+}
+
+static void
+cbe_skel_interface_print_inheritance_relayer(IDL_tree node, CBESkelInfo *current)
+{
+  IDL_tree curitem;
+  char *iface_id, *cur_id;
+
+  g_assert(current->tree);
+  g_assert(IDL_NODE_TYPE(node) == IDLN_INTERFACE);
+
+  if(node==current->tree) {
+    /* We don't inherit from ourself... */
+    return;
+  }
+
+  /* Print the operations defined for this interface, but in current's
+     namespace */
+
+  iface_id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(node).ident), "_", 0);
+  cur_id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(current->tree).ident), "_", 0);
+
+  for(curitem = IDL_INTERFACE(node).body;
+      curitem; curitem = IDL_LIST(curitem).next) {
+
+    cbe_skel_interface_print_relayer_internal(current, curitem, iface_id, cur_id);
+  }
+
+  free(iface_id);
+  free(cur_id);
+}
+
 void
 cbe_skel_interface_print_relayer(CBESkelInfo *ski)
 {
   /* XXX Prime candidate for optimization here. Is using gperf reasonable?
      In any case, there IS a better way ;-)
      -- Elliot */
-  /* XXX this won't work with inheritance. Fix. */
   char *id;
-  IDL_tree curitem, curdcl;
+  IDL_tree curitem;
 
   id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(ski->tree).ident), "_", 0);
-  fprintf(ski->of, "static ORBitSkeleton get_skel_%s(POA_%s *servant, GIOPRecvBuffer *_ORBIT_recv_buffer, gpointer *epv)\n", id, id);
+  fprintf(ski->of, "static ORBitSkeleton get_skel_%s(POA_%s *servant,\nGIOPRecvBuffer *_ORBIT_recv_buffer,\ngpointer *epv)\n", id, id);
   fprintf(ski->of, "{\n");
 
   for(curitem = IDL_INTERFACE(ski->tree).body;
       curitem; curitem = IDL_LIST(curitem).next) {
-
-    switch(IDL_NODE_TYPE(IDL_LIST(curitem).data)) {
-
-    case IDLN_OP_DCL:
-      fprintf(ski->of, "  if(!strcmp(_ORBIT_recv_buffer->message.u.request.operation, \"%s\")) {\n",
-	      IDL_IDENT(IDL_OP_DCL(IDL_LIST(curitem).data).ident).str);
-      fprintf(ski->of, "    *epv = servant->vepv->%s_epv->%s;\n", id,
-	      IDL_IDENT(IDL_OP_DCL(IDL_LIST(curitem).data).ident).str);
-      fprintf(ski->of,
-	      "    return (gpointer)skel_%s_%s;\n",
-	      id, IDL_IDENT(IDL_OP_DCL(IDL_LIST(curitem).data).ident).str);
-      fprintf(ski->of, "  } else");
-      break;
-
-    case IDLN_ATTR_DCL:
-      for(curdcl = IDL_ATTR_DCL(IDL_LIST(curitem).data).simple_declarations;
-	  curdcl; curdcl = IDL_LIST(curdcl).next) {
-
-	fprintf(ski->of, "  if(!strcmp(_ORBIT_recv_buffer->message.u.request.operation, \"_get_%s\")) {\n",
-		IDL_IDENT(IDL_LIST(curdcl).data).str);
-	fprintf(ski->of, "    *epv = servant->vepv->%s_epv->_get_%s;\n", id,
-		IDL_IDENT(IDL_LIST(curdcl).data).str);
-	fprintf(ski->of,
-		"    return (gpointer)skel_%s__get_%s;\n",
-		id,
-		IDL_IDENT(IDL_LIST(curdcl).data).str);
-	fprintf(ski->of, "  } else");
-
-	if(!IDL_ATTR_DCL(IDL_LIST(curitem).data).f_readonly) {
-	  fprintf(ski->of,
-		  "  if(!strcmp(_ORBIT_recv_buffer->message.u.request.operation, \"_set_%s\")) {\n",
-		  IDL_IDENT(IDL_LIST(curdcl).data).str);
-	  fprintf(ski->of, "    *epv = servant->vepv->%s_epv->_set_%s;\n", id,
-		IDL_IDENT(IDL_LIST(curdcl).data).str);
-	  fprintf(ski->of,
-		  "    return (gpointer)skel_%s__set_%s;\n",
-		  id, IDL_IDENT(IDL_LIST(curdcl).data).str);
-	  fprintf(ski->of, "  } else");
-	}
-      }
-      break;
-    default:
-    }
+    cbe_skel_interface_print_relayer_internal(ski, curitem, id, id);
   }
 
+  free(id);
+
+  IDL_tree_traverse_parents(ski->tree, (gpointer)cbe_skel_interface_print_inheritance_relayer, ski);
+
   fprintf(ski->of, "  {\n");
   fprintf(ski->of, "    g_error(\"Unknown operation\");\n");
   fprintf(ski->of, "  }\n");
   fprintf(ski->of, "  return NULL;\n");
 
 
+  fprintf(ski->of, "}\n\n");
+}
+
+static void
+cbe_skel_interface_print_initializer(IDL_tree node,
+				     CBESkelInfo *ski)
+{
+  char *id;
+
+  g_assert(IDL_NODE_TYPE(node) == IDLN_INTERFACE);
+
+  /* Print the operations defined for this interface, but in current's
+     namespace */
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(node).ident),
+			       "_", 0);
+
+  fprintf(ski->of, "obj->vepv[%s__classid] = servant->vepv->%s_epv;\n", id, id);
+
+  free(id);
+}
+
+static void
+cbe_skel_interface_print_objref_initializer(CBESkelInfo *ski)
+{
+  char *id;
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(ski->tree).ident), "_", 0);
+  fprintf(ski->of, "static void init_local_objref_%s(CORBA_Object obj, POA_%s *servant)\n", id, id);
+  fprintf(ski->of, "{\n");
+
+  free(id);
+
+  IDL_tree_traverse_parents(ski->tree,
+			    (gpointer)cbe_skel_interface_print_initializer,
+			    ski);
+
   fprintf(ski->of, "}\n");
 }
 
+static void
+cbe_skel_interface_print_inheritance_ops(IDL_tree node, CBESkelInfo *current)
+{
+  CBESkelInfo subski = *current;
+  IDL_tree curitem;
+  char *id;
+
+  g_assert(current->tree);
+  g_assert(IDL_NODE_TYPE(node) == IDLN_INTERFACE);
+
+  if(node==current->tree) {
+    /* We don't inherit from ourself... */
+    return;
+  }
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(node).ident), "_", 0);
+  fprintf(current->of, "/* From interface %s */\n", id);
+  free(id);
+
+  /* Print the operations defined for this interface, but in current's
+     namespace */
+
+  for(curitem = IDL_INTERFACE(node).body;
+      curitem; curitem = IDL_LIST(curitem).next) {
+
+    subski.tree = IDL_LIST(curitem).data;
+
+    switch(IDL_NODE_TYPE(IDL_LIST(curitem).data)) {
+    case IDLN_OP_DCL:
+      cbe_skel_do_inherited_op_dcl(&subski, current->tree);
+      break;
+    case IDLN_ATTR_DCL:
+      cbe_skel_do_inherited_attr_dcl(&subski, current->tree);
+      break;
+    default:
+    }
+  }
+}
+
 static void
 cbe_skel_do_interface(CBESkelInfo *ski)
 {
@@ -457,14 +629,24 @@ cbe_skel_do_interface(CBESkelInfo *ski)
 
   id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(ski->tree).ident), "_", 0);
 
+  subski.tree = IDL_INTERFACE(ski->tree).inheritance_spec;
+  if(subski.tree != NULL) {
+    fprintf(ski->of, "/* Begin inherited operations for %s */\n\n", id);
+    IDL_tree_traverse_parents(subski.tree, (gpointer)cbe_skel_interface_print_inheritance_ops, ski);
+    fprintf(ski->of, "/* End inherited operations for %s */\n\n", id);
+  }
+
   cbe_skel_interface_print_relayer(ski);
 
+  cbe_skel_interface_print_objref_initializer(ski);
+
   fprintf(ski->of,
-	  "void POA_%s__init(POA_%s *servant, CORBA_Environment *env)\n",
+	  "void POA_%s__init(POA_%s *servant,\nCORBA_Environment *env)\n",
 	  id, id);
   fprintf(ski->of, "{\n");
-  fprintf(ski->of, "  static PortableServer_ClassInfo class_info = {(gpointer)&get_skel_%s, \"IDL:%s:1.0\"};\n",
-	  id, id);
+  fprintf(ski->of,
+	  "  static const PortableServer_ClassInfo class_info = {(gpointer)&get_skel_%s, \"%s\", (gpointer)&init_local_objref_%s};\n",
+	  id, IDL_IDENT(IDL_INTERFACE(ski->tree).ident).repo_id, id);
 
   fprintf(ski->of,
 	  "  PortableServer_ServantBase__init(((PortableServer_ServantBase *)servant), env);\n");
@@ -477,12 +659,16 @@ cbe_skel_do_interface(CBESkelInfo *ski)
     free(id2);
   }
 
-  fprintf(ski->of, "  ORBIT_OBJECT_KEY(servant->_private)->class_info = &class_info;\n");
-  
+  fprintf(ski->of, "  ORBIT_OBJECT_KEY(servant->_private)->class_info = (PortableServer_ClassInfo*) &class_info;\n");
+
+  fprintf(ski->of,
+	  "if(!%s__classid)\n%s__classid = ORBit_register_class(&class_info);\n",
+	  id, id);
+
   fprintf(ski->of, "}\n\n");
 
   fprintf(ski->of,
-	  "void POA_%s__fini(POA_%s *servant, CORBA_Environment *env)\n",
+	  "void POA_%s__fini(POA_%s *servant,\nCORBA_Environment *env)\n",
 	  id, id);
   fprintf(ski->of, "{\n");
   if(IDL_INTERFACE(ski->tree).inheritance_spec)
@@ -540,39 +726,43 @@ cbe_skel_do_none(CBESkelInfo *ski)
 }
 
 static void
-cbe_skel_do_op_dcl(CBESkelInfo *ski)
+cbe_skel_op_dcl_print_call_param(CBESkelInfo *ski)
+{
+  int i, n;
+
+  n = orbit_cbe_param_numptrs(ski->tree,
+			      cbe_attr_to_paramrole(IDL_PARAM_DCL(ski->tree).attr));
+
+  n = n && n; /* just one & */
+  for(i = 0; i < n; i++)
+    fprintf(ski->of, "&(");
+
+  fprintf(ski->of, "%s",
+	  IDL_IDENT(IDL_PARAM_DCL(ski->tree).simple_declarator).str);
+
+  for(i = 0; i < n; i++)
+    fprintf(ski->of, ")");
+}
+
+static void
+cbe_skel_do_op_dcl_shared(CBESkelInfo *ski)
 {
   /* XXX we don't handle contexts here yet - fix that */
-  char *id, *id2;
   IDL_tree curitem;
-  int level, i, n;
-  CBEMarshalInfo mi = {ski->of, NULL, NULL, NULL, NULL};
-  CBEDemarshalInfo dmi = {ski->of, NULL, NULL, NULL, NULL, TRUE, FALSE};
   GString *tmpstr = g_string_new(NULL);
-
-  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_OP_DCL(ski->tree).ident),
-			       "_", 0);
-
-  if(IDL_OP_DCL(ski->tree).op_type_spec && IDL_OP_DCL(ski->tree).f_oneway) {
-    g_error("[%s] You cannot have a return value from a oneway operation!\n", id);
-  }
-  curitem = IDL_get_parent_node(ski->tree, IDLN_INTERFACE, &level);
-
-  g_assert(curitem);
-  id2 = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(curitem).ident), "_", 0);
-
-  fprintf(ski->of, "\nvoid\nskel_%s(POA_%s *_ORBIT_servant, GIOPRecvBuffer *_ORBIT_recv_buffer, CORBA_Environment *ev,\n", id, id2);
+  CBESkelInfo subski = *ski;
+  char *id;
 
   orbit_cbe_write_typespec(ski->of, IDL_OP_DCL(ski->tree).op_type_spec);
   orbit_cbe_param_printptrs(ski->of, IDL_OP_DCL(ski->tree).op_type_spec,
 			    DATA_RETURN);
 
-  fprintf(ski->of, " (*%s)(PortableServer_Servant servant, ",
+  fprintf(ski->of, " (*_impl_%s)(PortableServer_Servant servant,\n",
 	  IDL_IDENT(IDL_OP_DCL(ski->tree).ident).str);
 
   for(curitem = IDL_OP_DCL(ski->tree).parameter_dcls; curitem; curitem = IDL_LIST(curitem).next) {
     cbe_print_param_dcl(ski->of, IDL_LIST(curitem).data);
-    fprintf(ski->of, ", ");
+    fprintf(ski->of, ",\n");
   }
   fprintf(ski->of, "CORBA_Environment *ev))\n");
 
@@ -583,72 +773,55 @@ cbe_skel_do_op_dcl(CBESkelInfo *ski)
   for(curitem = IDL_OP_DCL(ski->tree).parameter_dcls; curitem;
       curitem = IDL_LIST(curitem).next) {
     fprintf(ski->of, "  ");
-    cbe_print_var_dcl(ski->of, IDL_LIST(curitem).data);
+    cbe_print_var_dcl(ski->of, IDL_LIST(curitem).data, TRUE);
+    fprintf(ski->of, ";\n");
+  }
+
+  if(IDL_OP_DCL(ski->tree).op_type_spec) {
+    fprintf(ski->of, "  ");
+    cbe_print_var_dcl(ski->of, IDL_OP_DCL(ski->tree).op_type_spec, TRUE);
     fprintf(ski->of, ";\n");
   }
 
   if(!IDL_OP_DCL(ski->tree).f_oneway)
     fprintf(ski->of, "  GIOPSendBuffer *_ORBIT_send_buffer;\n");
 
-  if(IDL_OP_DCL(ski->tree).op_type_spec) {
-    fprintf(ski->of, "  ");
-    cbe_print_param_dcl(ski->of, IDL_OP_DCL(ski->tree).op_type_spec);
-    fprintf(ski->of, ";\n");
+  if(IDL_OP_DCL(ski->tree).raises_expr) {
+    fprintf(ski->of, "static const ORBit_exception_marshal_info _ORBIT_user_exceptions[] = {");
+    for(curitem = IDL_OP_DCL(ski->tree).raises_expr; curitem;
+	curitem = IDL_LIST(curitem).next) {
+      IDL_tree curnode = IDL_LIST(curitem).data;
+
+      id = orbit_cbe_get_typename(curnode);
+      fprintf(ski->of, "{(CORBA_TypeCode)&TC_%s_struct, (gpointer)%s_marshal},", id, id);
+      g_free(id);
+    }
+    fprintf(ski->of, "{CORBA_OBJECT_NIL, NULL}};\n");
   }
+
   fprintf(ski->of, "\n");
 
-  if(cbe_op_dcl_nparams(ski->tree, DATA_INOUT|DATA_IN) > 0) {
+  if(cbe_op_dcl_nparams(ski->tree, DATA_INOUT|DATA_IN) > 0)
+    cbe_skel_op_params_demarshal(ski);
 
-    fprintf(ski->of, "  if(giop_msg_conversion_needed(GIOP_MESSAGE_BUFFER(_ORBIT_recv_buffer))) {\n");
-
-    do {
-      for(curitem = IDL_OP_DCL(ski->tree).parameter_dcls;
-	  curitem; curitem = IDL_LIST(curitem).next) {
-
-	if(IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_IN
-	   || IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_INOUT) {
-	
-	  if(IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_IN)
-	    n = orbit_cbe_param_numptrs(mi.param, DATA_IN);
-	  else
-	    n = orbit_cbe_param_numptrs(mi.param, DATA_INOUT) /* only getting, not setting */ - 1;
-	
-	  dmi.previous_param = dmi.param;
-	  dmi.param = IDL_LIST(curitem).data;
-	
-	  g_string_assign(tmpstr, "");
-	  for(i = 0; i < n; i++)
-	    g_string_append_c(tmpstr, '*');
-	  g_string_sprintfa(tmpstr, "%s",
-			    IDL_IDENT(IDL_PARAM_DCL(dmi.param).simple_declarator).str);
-	  dmi.param_name = tmpstr->str;
-	  fprintf(ski->of, "\n\n  /* demarshal parameter %s */\n", dmi.param_name);
-	  cbe_output_demarshaller(dmi);
-	}
-      }
-      if(dmi.byteswap_version)
-	fprintf(ski->of, "  } else {\n");
-
-    } while(dmi.byteswap_version-- > 0);
-
-    fprintf(ski->of, "  }\n\n");
-
-  }
-  
+  /* Make things happen - call the implementation function */
   if(IDL_OP_DCL(ski->tree).op_type_spec)
     fprintf(ski->of, "  _ORBIT_retval = ");
-  fprintf(ski->of, "%s(_ORBIT_servant, ",
+
+  fprintf(ski->of, "_impl_%s(_ORBIT_servant, ",
 	  IDL_IDENT(IDL_OP_DCL(ski->tree).ident).str);
 
   for(curitem = IDL_OP_DCL(ski->tree).parameter_dcls; curitem;
       curitem = IDL_LIST(curitem).next) {
-    if(IDL_PARAM_DCL(IDL_LIST(curitem).data).attr != IDL_PARAM_IN)
-      fprintf(ski->of, "&");
-    fprintf(ski->of, "%s,",
-	    IDL_IDENT(IDL_PARAM_DCL(IDL_LIST(curitem).data).simple_declarator).str);
+
+    subski.tree = IDL_LIST(curitem).data;
+    cbe_skel_op_dcl_print_call_param(&subski);
+
+    fprintf(ski->of, ",\n");
   }
   fprintf(ski->of, " ev);\n");
 
+  /* Now marshal any return params and send it out over the wire */
   if(!IDL_OP_DCL(ski->tree).f_oneway) {
     fprintf(ski->of,
 	    "  _ORBIT_send_buffer = \n");
@@ -657,74 +830,342 @@ cbe_skel_do_op_dcl(CBESkelInfo *ski)
     fprintf(ski->of,
 	    "	 _ORBIT_recv_buffer->message.u.request.request_id,\n");
     fprintf(ski->of,
-	    "	 GIOP_NO_EXCEPTION);\n\n");
+	    "	 ev->_major);\n\n");
 
-
-    if(cbe_op_dcl_nparams(ski->tree, DATA_INOUT|DATA_OUT|DATA_RETURN) > 0) {
-
-      mi.param = NULL;
-
-      if(IDL_OP_DCL(ski->tree).op_type_spec) {
-	fprintf(ski->of, "  /* marshal return value */\n");
-	mi.param = IDL_OP_DCL(ski->tree).op_type_spec;
-	mi.param_name = "_ORBIT_retval";
-	cbe_output_marshaller(mi);
-      }
-
-      for(curitem = IDL_OP_DCL(ski->tree).parameter_dcls;
-	  curitem; curitem = IDL_LIST(curitem).next) {
-	
-	if(IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_OUT
-	   || IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_INOUT) {
-
-	  mi.previous_param = mi.param;
-	  mi.param = IDL_LIST(curitem).data;
-
-	  if(IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_OUT)
-	    n = orbit_cbe_param_numptrs(mi.param, DATA_OUT);
-	  else
-	    n = orbit_cbe_param_numptrs(mi.param, DATA_INOUT); /* Clarify */
-
-	  g_string_assign(tmpstr, "");
-#if 0
-	  for(i = 0; i < n; i++)
-	    g_string_append_c(tmpstr, '*');
-#endif
-	  g_string_sprintfa(tmpstr, "%s",
-			    IDL_IDENT(IDL_PARAM_DCL(mi.param).simple_declarator).str);
-	  mi.param_name = tmpstr->str;
-
-	  fprintf(ski->of, "  /* marshal parameter %s */\n", mi.param_name);
-	  cbe_output_marshaller(mi);
-	}
-      }
-    }
+    fprintf(ski->of, "if(ev->_major == CORBA_NO_EXCEPTION) {\n");
+    cbe_skel_op_params_marshal(ski);
+    fprintf(ski->of, "} else if(ev->_major == CORBA_SYSTEM_EXCEPTION)\n");
+    fprintf(ski->of, " ORBit_send_system_exception(_ORBIT_send_buffer, ev);");
+    fprintf(ski->of, "else if(ev->_major == CORBA_USER_EXCEPTION)\n");
+    fprintf(ski->of, " ORBit_send_user_exception(_ORBIT_send_buffer, ev, %s);",
+	    IDL_OP_DCL(ski->tree).raises_expr?"_ORBIT_user_exceptions":"NULL");
 
     fprintf(ski->of, "  giop_send_buffer_write(_ORBIT_send_buffer);\n");
     fprintf(ski->of, "  giop_send_buffer_unuse(_ORBIT_send_buffer);\n");
 
-    if(IDL_OP_DCL(ski->tree).op_type_spec
-       && IDL_NODE_TYPE(IDL_OP_DCL(ski->tree).op_type_spec) == IDLN_TYPE_STRING) {
-      fprintf(ski->of, "  CORBA_free(_ORBIT_retval);\n");
-    }
-
-    for(curitem = IDL_OP_DCL(ski->tree).parameter_dcls;
-	curitem; curitem = IDL_LIST(curitem).next) {
-      if(IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_OUT
-	 || IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_INOUT)
-	switch(IDL_NODE_TYPE(IDL_PARAM_DCL(IDL_LIST(curitem).data).param_type_spec)) {
-	case IDLN_TYPE_STRING:
-	  fprintf(ski->of, "  CORBA_free(%s);\n",
-		  IDL_IDENT(IDL_PARAM_DCL(IDL_LIST(curitem).data).simple_declarator).str);
-	  break;
-	default:
-	}
-    }
+    cbe_skel_op_params_free(ski);
   }
 
   fprintf(ski->of, "}\n\n");
   g_string_free(tmpstr, TRUE);
+}
+
+static void
+cbe_skel_do_op_dcl(CBESkelInfo *ski)
+{
+  /* XXX we don't handle contexts here yet - fix that */
+  char *id, *id2;
+  IDL_tree curitem;
+  int level;
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_OP_DCL(ski->tree).ident),
+			       "_", 0);
+
+  if(IDL_OP_DCL(ski->tree).op_type_spec && IDL_OP_DCL(ski->tree).f_oneway) {
+    g_error("[%s] You cannot have a return value from a oneway operation!", id);
+  }
+  curitem = IDL_get_parent_node(ski->tree, IDLN_INTERFACE, &level);
+
+  g_assert(curitem);
+  id2 = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(curitem).ident), "_", 0);
+
+  fprintf(ski->of, "\nvoid\nskel_%s(POA_%s *_ORBIT_servant,\nGIOPRecvBuffer *_ORBIT_recv_buffer,\nCORBA_Environment *ev,\n", id, id2);
   free(id); free(id2);
+
+  cbe_skel_do_op_dcl_shared(ski);
+}
+
+/* FIXME This function should check for redefined inherited ops, it doesn't
+   currently - RHP */
+static void
+cbe_skel_do_inherited_op_dcl(CBESkelInfo *ski, IDL_tree current_interface)
+{
+  char *id;
+  IDL_tree ident;
+
+  id = IDL_ns_ident_to_qstring(IDL_IDENT_TO_NS(IDL_INTERFACE(current_interface).ident), "_", 0);
+  ident = IDL_OP_DCL(ski->tree).ident;
+  g_assert(ident);
+
+  if(IDL_OP_DCL(ski->tree).op_type_spec && IDL_OP_DCL(ski->tree).f_oneway) {
+    g_error("[%s_%s] You cannot have a return value from a oneway operation!", id, IDL_IDENT(ident).str);
+  }
+
+  fprintf(ski->of, "\nvoid\nskel_%s_%s(POA_%s *_ORBIT_servant,\nGIOPRecvBuffer *_ORBIT_recv_buffer,\nCORBA_Environment *ev,\n", id, IDL_IDENT(ident).str, id);
+  free(id);
+
+  cbe_skel_do_op_dcl_shared(ski);
+}
+
+static void
+cbe_skel_op_params_marshal(CBESkelInfo *ski)
+{
+  CBEMarshalInfo mi = {ski->of, NULL, NULL, NULL, NULL};
+  IDL_tree curitem;
+  GString *tmpstr;
+  int n, i;
+
+  if(cbe_op_dcl_nparams(ski->tree, DATA_INOUT|DATA_OUT|DATA_RETURN) <= 0)
+    return;
+
+  tmpstr = g_string_new(NULL);
+  mi.param = mi.previous_param = NULL;
+
+  if(IDL_OP_DCL(ski->tree).op_type_spec) {
+    fprintf(ski->of, "  /* marshal return value */\n");
+    mi.param = IDL_OP_DCL(ski->tree).op_type_spec;
+
+    n = orbit_cbe_param_numptrs(mi.param, DATA_RETURN);
+
+    g_string_assign(tmpstr, "");
+
+    for(i = 0; i < n; i++)
+      g_string_append_c(tmpstr, '*');
+
+    g_string_append(tmpstr, "_ORBIT_retval");
+
+    mi.param_name = tmpstr->str;
+    cbe_output_marshaller(&mi);
+  }
+
+  for(curitem = IDL_OP_DCL(ski->tree).parameter_dcls;
+      curitem; curitem = IDL_LIST(curitem).next) {
+	
+    if(IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_OUT
+       || IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_INOUT) {
+
+      mi.param = IDL_LIST(curitem).data;
+
+      n = orbit_cbe_param_numptrs(mi.param,
+				  cbe_attr_to_paramrole(IDL_PARAM_DCL(mi.param).attr)) - 1;
+
+      if(cbe_type_is_fixed_length(mi.param))
+	n = 0;
+
+      g_string_assign(tmpstr, "");
+
+      for(i = 0; i < n; i++)
+	g_string_append_c(tmpstr, '*');
+
+      g_string_sprintfa(tmpstr, "%s",
+			IDL_IDENT(IDL_PARAM_DCL(mi.param).simple_declarator).str);
+      mi.param_name = tmpstr->str;
+
+      fprintf(ski->of, "  /* marshal parameter %s */\n", mi.param_name);
+      cbe_output_marshaller(&mi);
+    }
+  }  
+
+  g_string_free(tmpstr, TRUE);
+}
+
+static void
+cbe_skel_param_subfree(CBESkelInfo *ski)
+{
+  char *id, *varname;
+
+  if(IDL_NODE_TYPE(ski->tree) != IDLN_PARAM_DCL) {
+    id = orbit_cbe_get_typename(ski->tree);
+    varname = "_ORBIT_retval";
+  } else {
+    id = orbit_cbe_get_typename(IDL_PARAM_DCL(ski->tree).param_type_spec);
+    varname = IDL_IDENT(IDL_PARAM_DCL(ski->tree).simple_declarator).str;
+  }
+
+  fprintf(ski->of, "%s__free(&%s, NULL, %s);\n",
+	  id, varname, ski->free_internal);
+  free(id);
+}
+
+static void
+cbe_skel_op_retval_free(CBESkelInfo *ski)
+{
+  IDL_tree ts;
+
+  ts = cbe_get_typespec(ski->tree);
+
+  switch(IDL_NODE_TYPE(ts)) {
+  case IDLN_TYPE_UNION:
+  case IDLN_TYPE_STRUCT:
+    if(cbe_type_is_fixed_length(ts))
+      return;
+  case IDLN_TYPE_SEQUENCE:
+  case IDLN_TYPE_STRING:
+  case IDLN_TYPE_ARRAY:
+  case IDLN_TYPE_ANY:
+    break;
+  case IDLN_TYPE_OBJECT:
+  case IDLN_INTERFACE:
+  case IDLN_FORWARD_DCL:
+    fprintf(ski->of, "if(ev->_major == CORBA_NO_EXCEPTION) CORBA_Object_release(_ORBIT_retval, ev);\n");
+  default:
+    return;
+  }
+
+  fprintf(ski->of, "if(ev->_major == CORBA_NO_EXCEPTION) CORBA_free(_ORBIT_retval);\n");
+}
+
+static gboolean
+cbe_skel_op_param_has_sequence(IDL_tree ts)
+{
+  gboolean has_seq = FALSE, subhas;
+  IDL_tree curitem, sn;
+
+  ts = cbe_get_typespec(ts);
+
+  switch(IDL_NODE_TYPE(ts)) {
+  case IDLN_TYPE_UNION:
+    for(curitem = IDL_TYPE_UNION(ts).switch_body; curitem;
+	curitem = IDL_LIST(curitem).next) {
+      sn = IDL_MEMBER(IDL_CASE_STMT(IDL_LIST(curitem).data).element_spec).type_spec;
+      subhas = cbe_skel_op_param_has_sequence(sn);
+      has_seq = has_seq || subhas;
+    }
+    return has_seq;
+    break;
+  case IDLN_TYPE_STRUCT:
+    for(curitem = IDL_TYPE_STRUCT(ts).member_list; curitem;
+	curitem = IDL_LIST(curitem).next) {
+      sn = IDL_MEMBER(IDL_LIST(curitem).data).type_spec;
+      subhas = cbe_skel_op_param_has_sequence(sn);
+      has_seq = has_seq || subhas;
+    }
+    return has_seq;
+    break;
+  case IDLN_TYPE_ARRAY:
+    return cbe_skel_op_param_has_sequence(IDL_TYPE_DCL(IDL_get_parent_node(ts, IDLN_TYPE_DCL, NULL)).type_spec);
+    break;
+  case IDLN_TYPE_SEQUENCE:
+    return TRUE;
+  default:
+    return FALSE;
+  }
+}
+
+static void
+cbe_skel_op_param_free(CBESkelInfo *ski)
+{
+  IDL_tree ts;
+
+  ts = cbe_get_typespec(ski->tree);
+
+  switch(IDL_PARAM_DCL(ski->tree).attr) {
+  case IDL_PARAM_IN:
+    if(cbe_type_is_fixed_length(ts))
+      return;
+    ski->free_internal = "CORBA_FALSE";
+    switch(IDL_NODE_TYPE(ts)) {
+    case IDLN_TYPE_SEQUENCE:
+    case IDLN_TYPE_UNION:
+    case IDLN_TYPE_STRUCT:
+    case IDLN_TYPE_ARRAY:
+    case IDLN_TYPE_ANY:
+      cbe_skel_param_subfree(ski);
+      break;
+    case IDLN_TYPE_OBJECT:
+    case IDLN_INTERFACE:
+    case IDLN_FORWARD_DCL:
+      fprintf(ski->of, "CORBA_Object_release(%s, ev);\n",
+	      IDL_IDENT(IDL_PARAM_DCL(ski->tree).simple_declarator).str);
+      break;
+    default:
+    }
+    break;
+  case IDL_PARAM_OUT:
+    if(cbe_type_is_fixed_length(ts))
+      return;
+
+    fprintf(ski->of, "if(ev->_major == CORBA_NO_EXCEPTION)");
+    switch(IDL_NODE_TYPE(ts)) {
+    case IDLN_TYPE_OBJECT:
+    case IDLN_INTERFACE:
+      fprintf(ski->of, "CORBA_Object_release(%s, ev);\n",
+	      IDL_IDENT(IDL_PARAM_DCL(ski->tree).simple_declarator).str);
+      break;
+    default:
+      fprintf(ski->of, "CORBA_free(%s);\n",
+	      IDL_IDENT(IDL_PARAM_DCL(ski->tree).simple_declarator).str);
+    }
+    break;
+  case IDL_PARAM_INOUT:
+    if(cbe_type_is_fixed_length(ts))
+      return;
+    switch(IDL_NODE_TYPE(ts)) {
+    case IDLN_TYPE_OBJECT:
+    case IDLN_INTERFACE:
+      fprintf(ski->of, "CORBA_Object_release(%s, ev);\n",
+	      IDL_IDENT(IDL_PARAM_DCL(ski->tree).simple_declarator).str);
+      break;
+    default:
+      cbe_skel_param_subfree(ski);
+    }
+    break;
+  }
+
+}
+
+static void
+cbe_skel_op_params_free(CBESkelInfo *ski)
+{
+  IDL_tree curitem;
+  CBESkelInfo subski = *ski;
+  
+  if(IDL_OP_DCL(ski->tree).op_type_spec) {
+    subski.tree = IDL_OP_DCL(ski->tree).op_type_spec;
+    cbe_skel_op_retval_free(&subski);
+  }
+  
+  for(curitem = IDL_OP_DCL(ski->tree).parameter_dcls;
+      curitem; curitem = IDL_LIST(curitem).next) {
+    subski.tree = IDL_LIST(curitem).data;
+    subski.free_internal = "CORBA_TRUE";
+    cbe_skel_op_param_free(&subski);
+  }
+}
+
+static void
+cbe_skel_op_params_demarshal(CBESkelInfo *ski)
+{
+  CBEDemarshalInfo dmi = {ski->of, NULL, NULL, NULL, NULL, TRUE, FALSE,
+			  "(((ORBit_ObjectKey *)_ORBIT_servant->_private)->object->orb)"};
+  IDL_tree curitem;
+  GString *tmpstr = g_string_new(NULL);
+
+  fprintf(ski->of,
+	  "  if(giop_msg_conversion_needed(GIOP_MESSAGE_BUFFER(_ORBIT_recv_buffer))) {\n");
+
+  do {
+    dmi.previous_param = NULL;
+
+    for(curitem = IDL_OP_DCL(ski->tree).parameter_dcls;
+	curitem; curitem = IDL_LIST(curitem).next) {
+
+      if(IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_IN
+	 || IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_INOUT) {
+
+	if(IDL_PARAM_DCL(IDL_LIST(curitem).data).attr == IDL_PARAM_INOUT)
+	  dmi.allocate_memory = TRUE;
+	else
+	  dmi.allocate_memory = FALSE;
+
+	dmi.param = IDL_LIST(curitem).data;
+
+	g_string_assign(tmpstr, "");
+	g_string_sprintfa(tmpstr, "%s",
+			  IDL_IDENT(IDL_PARAM_DCL(dmi.param).simple_declarator).str);
+	dmi.param_name = tmpstr->str;
+	fprintf(ski->of, "\n\n  /* demarshal parameter %s */\n", dmi.param_name);
+	cbe_output_demarshaller(&dmi);
+      }
+    }
+    if(dmi.byteswap_version)
+      fprintf(ski->of, "  } else {\n");
+
+  } while(dmi.byteswap_version-- > 0);
+
+  fprintf(ski->of, "  }\n\n");
+
+  g_string_free(tmpstr, TRUE);
 }
 
 static void
@@ -742,121 +1183,121 @@ cbe_skel_do_string(CBESkelInfo *ski)
 static void
 cbe_skel_do_type_any(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_array(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_boolean(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_char(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_dcl(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_enum(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_fixed(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_float(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_integer(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_object(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_octet(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_sequence(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_string(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_struct(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_union(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_wide_char(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_type_wide_string(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_unaryop(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_wide_char(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 static void
 cbe_skel_do_wide_string(CBESkelInfo *ski)
 {
-  g_assert(!"Not yet implemented");
+  //g_assert(!"Not yet implemented");
 }
 
 /* Dead code repo */
