@@ -5,9 +5,9 @@
 #ifndef ALLOCATORS_H
 #define ALLOCATORS_H 1
 
-#include "orbit.h"
+#include <orb/orbit.h>
 
-#include "allocator-defs.h"
+#include <orb/allocator-defs.h>
 
 #define ORBIT_CHUNK_ALLOC(typename) \
 ORBit_chunk_alloc(typename##_allocator, LOCK_NAME(typename##_allocator))
@@ -30,7 +30,8 @@ void ORBit_chunk_free(GMemChunk *chunk,
 #define MEMINFO_TO_PTR(x) ((gpointer)((x) + 1))
 
 typedef gpointer (*ORBit_free_childvals)(gpointer mem,
-					 gpointer func_data);
+					 gpointer func_data,
+					 CORBA_boolean free_strings);
 
 typedef struct {
 #ifdef DEBUG
@@ -45,7 +46,16 @@ typedef struct {
 gpointer ORBit_alloc(size_t block_size,
 		     ORBit_free_childvals freefunc,
 		     gpointer func_data);
+gpointer ORBit_alloc_2(size_t block_size,
+		       ORBit_free_childvals freefunc,
+		       gpointer func_data,
+		       size_t before_size);
 
-void ORBit_free(gpointer mem);
+void ORBit_free(gpointer mem, CORBA_boolean free_strings);
+
+/* internal stuff */
+gpointer ORBit_free_via_TypeCode(gpointer mem,
+				 gpointer tcp,
+				 gboolean free_strings);
 
 #endif /* ALLOCATORS_H */
