@@ -24,13 +24,12 @@
  */
 
 #include "config.h"
-
+#include "../IIOP/iiop-endianP.h"
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
 #include <assert.h>
 
-#include "ORBitutil/machine_endian.h"
 #include "orbit.h"
 
 #define CDR_GROW_AMOUNT 128
@@ -262,6 +261,7 @@ static void CDR_buffer_put8(CDR_Codec *codec, void *datum)
 	codec->buffer[codec->wptr++]=((CORBA_octet *)datum)[7];
 }
 
+#if 0
 static CORBA_boolean CDR_buffer_get8(CDR_Codec *codec, void *dest)
 {
 	unsigned long align;
@@ -290,6 +290,7 @@ static CORBA_boolean CDR_buffer_get8(CDR_Codec *codec, void *dest)
 
 	return(CORBA_TRUE);
 }
+#endif
 
 static void CDR_buffer_put16(CDR_Codec *codec, void *datum)
 {
@@ -332,6 +333,7 @@ static void CDR_buffer_put16(CDR_Codec *codec, void *datum)
 	codec->buffer[codec->wptr++]=((CORBA_octet *)datum)[15];
 }
 
+#if 0
 static CORBA_boolean CDR_buffer_get16(CDR_Codec *codec, void *dest)
 {
 	unsigned long align;
@@ -368,6 +370,7 @@ static CORBA_boolean CDR_buffer_get16(CDR_Codec *codec, void *dest)
 
 	return(CORBA_TRUE);
 }
+#endif
 
 static void CDR_swap2(void *d, void *s)
 {
@@ -689,17 +692,18 @@ CDR_Codec *CDR_codec_init(void)
 	CDR_Codec *new;
 
 	new=g_new0(CDR_Codec, 1);
+	new->release_buffer = CORBA_TRUE;
+	new->host_endian = FLAG_ENDIANNESS;
+
 	if(new==NULL) {
 		ORBit_Trace(TraceMod_CDR, TraceLevel_Error, "CDR_codec_init: malloc error creating codec\n");
 		return(NULL);
 	}
 
-#if defined MACHINE_IS_LITTLE_ENDIAN
-	new->host_endian=LittleEndian;
-#elif defined MACHINE_IS_BIG_ENDIAN
+#ifdef WORDS_BIGENDIAN
 	new->host_endian=BigEndian;
 #else
-# error Cant determine host byte order
+	new->host_endian=LittleEndian;
 #endif
 
 	return(new);
@@ -709,7 +713,8 @@ void CDR_codec_free(CDR_Codec *codec)
 {
 	g_assert(codec!=NULL);
 
-	g_free(codec->buffer);
+	if(codec->release_buffer)
+		g_free(codec->buffer);
 
 	g_free(codec);
 }
