@@ -29,7 +29,6 @@
 #include <ORBitutil/basic_types.h>
 
 #include "corba_env.h"
-#include "corba_ir.h"
 #include "corba_orb.h"
 #include "corba_sequences.h"
 
@@ -39,6 +38,10 @@ typedef struct CORBA_Object_struct *CORBA_Object;
 
 #define CORBA_OBJECT(x) ((CORBA_Object)(x))
 
+/* Used for internal stuff mostly, but also good if you want to store
+   a hash of objects */
+gboolean g_CORBA_Object_equal(CORBA_Object obj1, CORBA_Object obj2);
+guint g_CORBA_Object_hash(CORBA_Object obj);
 
 #endif /* !_ORBIT_CORBA_OBJECT_H_ */
 
