@@ -42,35 +42,35 @@ struct SysExInfo {
 
 static const struct SysExInfo exception_table[] = {
 	{NULL, 0},
-	{"CORBA_UNKNOWN", 1},
-	{"CORBA_BAD_PARAM", 2},
-	{"CORBA_NO_MEMORY", 3},
-	{"CORBA_IMP_LIMIT", 4},
-	{"CORBA_COMM_FAILURE", 5},
-	{"CORBA_INV_OBJREF", 6},
-	{"CORBA_NO_PERMISSION", 7},
-	{"CORBA_INTERNAL", 8},
-	{"CORBA_MARSHAL", 9},
-	{"CORBA_INITIALIZE", 10},
-	{"CORBA_NO_IMPLEMENT", 11},
-	{"CORBA_BAD_TYPECODE", 12},
-	{"CORBA_BAD_OPERATION", 13},
-	{"CORBA_NO_RESOURCES", 14},
-	{"CORBA_NO_RESPONSE", 15},
-	{"CORBA_PERSIST_STORE", 16},
-	{"CORBA_BAD_INV_ORDER", 17},
-	{"CORBA_TRANSIENT", 18},
-	{"CORBA_FREE_MEM", 19},
-	{"CORBA_INV_IDENT", 20},
-	{"CORBA_INV_FLAG", 21},
-	{"CORBA_INTF_REPOS", 22},
-	{"CORBA_BAD_CONTEXT", 23},
-	{"CORBA_OBJ_ADAPTER", 24},
-	{"CORBA_DATA_CONVERSION", 25},
-	{"CORBA_OBJECT_NOT_EXIST", 26},
-	{"CORBA_TRANSACTION_REQUIRED", 27},
-	{"CORBA_TRANSACTION_ROLLEDBACK", 28},
-	{"CORBA_INVALID_TRANSACTION", 29},
+	{"IDL:CORBA/UNKNOWN:1.0", 1},
+	{"IDL:CORBA/BAD_PARAM:1.0", 2},
+	{"IDL:CORBA/NO_MEMORY:1.0", 3},
+	{"IDL:CORBA/IMP_LIMIT:1.0", 4},
+	{"IDL:CORBA/COMM_FAILURE:1.0", 5},
+	{"IDL:CORBA/INV_OBJREF:1.0", 6},
+	{"IDL:CORBA/NO_PERMISSION:1.0", 7},
+	{"IDL:CORBA/INTERNAL:1.0", 8},
+	{"IDL:CORBA/MARSHAL:1.0", 9},
+	{"IDL:CORBA/INITIALIZE:1.0", 10},
+	{"IDL:CORBA/NO_IMPLEMENT:1.0", 11},
+	{"IDL:CORBA/BAD_TYPECODE:1.0", 12},
+	{"IDL:CORBA/BAD_OPERATION:1.0", 13},
+	{"IDL:CORBA/NO_RESOURCES:1.0", 14},
+	{"IDL:CORBA/NO_RESPONSE:1.0", 15},
+	{"IDL:CORBA/PERSIST_STORE:1.0", 16},
+	{"IDL:CORBA/BAD_INV_ORDER:1.0", 17},
+	{"IDL:CORBA/TRANSIENT:1.0", 18},
+	{"IDL:CORBA/FREE_MEM:1.0", 19},
+	{"IDL:CORBA/INV_IDENT:1.0", 20},
+	{"IDL:CORBA/INV_FLAG:1.0", 21},
+	{"IDL:CORBA/INTF_REPOS:1.0", 22},
+	{"IDL:CORBA/BAD_CONTEXT:1.0", 23},
+	{"IDL:CORBA/OBJ_ADAPTER:1.0", 24},
+	{"IDL:CORBA/DATA_CONVERSION:1.0", 25},
+	{"IDL:CORBA/OBJECT_NOT_EXIST:1.0", 26},
+	{"IDL:CORBA/TRANSACTION_REQUIRED:1.0", 27},
+	{"IDL:CORBA/TRANSACTION_ROLLEDBACK:1.0", 28},
+	{"IDL:CORBA/INVALID_TRANSACTION:1.0", 29},
 	{NULL,0}
 };
 
@@ -126,7 +126,7 @@ void CORBA_exception_set_system(CORBA_Environment *ev, CORBA_unsigned_long minor
 
 		/* XXX what should the repo ID be? */
 		CORBA_exception_set(ev, CORBA_SYSTEM_EXCEPTION,
-				    (gchar *)exception_table[ev->_major].repo_id,
+				    (gchar *)exception_table[minor].repo_id,
 				    new);
 	}
 }
@@ -216,8 +216,11 @@ void ORBit_handle_exception(GIOPRecvBuffer *rb, CORBA_Environment *ev,
 	rb->decoder(&len, rb->cur, sizeof(len));
 	rb->cur += sizeof(len);
 
-	my_repoid = rb->cur;
-	rb->cur += len;
+	if(len) {
+		my_repoid = rb->cur;
+		rb->cur += len;
+	} else
+		my_repoid = NULL;
 
 	if(rb->message.u.reply.reply_status == CORBA_SYSTEM_EXCEPTION) {
 		ev->_major = CORBA_SYSTEM_EXCEPTION;
