@@ -59,9 +59,9 @@ void CORBA_any_set_release(CORBA_any *any, CORBA_boolean flag)
 	g_assert(any!=NULL);
 
 	if(flag==CORBA_TRUE) {
-		any->_flags |= CORBA_ANYFLAGS_RELEASE;
+		any->_release |= CORBA_ANYFLAGS_RELEASE;
 	} else {
-		any->_flags &= ~CORBA_ANYFLAGS_RELEASE;
+		any->_release &= ~CORBA_ANYFLAGS_RELEASE;
 	}
 
 }
@@ -70,7 +70,7 @@ CORBA_boolean CORBA_any_get_release(CORBA_any *any)
 {
 	g_assert(any!=NULL);
 
-	if(any->_flags & CORBA_ANYFLAGS_RELEASE)
+	if(any->_release & CORBA_ANYFLAGS_RELEASE)
 		return(CORBA_TRUE);
 	else
 		return(CORBA_FALSE);
@@ -85,9 +85,9 @@ void CORBA_sequence_set_release(void *seq, CORBA_boolean flag)
 	sequence=(struct CORBA_Sequence_type *)seq;
 
 	if(flag==CORBA_TRUE) {
-		sequence->_flags |= CORBA_ANYFLAGS_RELEASE;
+		sequence->_release |= CORBA_ANYFLAGS_RELEASE;
 	} else {
-		sequence->_flags &= ~CORBA_ANYFLAGS_RELEASE;
+		sequence->_release &= ~CORBA_ANYFLAGS_RELEASE;
 	}
 }
 
@@ -99,7 +99,7 @@ CORBA_boolean CORBA_sequence_get_release(void *seq)
 
 	sequence=(struct CORBA_Sequence_type *)seq;
 
-	if(sequence->_flags & CORBA_ANYFLAGS_RELEASE)
+	if(sequence->_release & CORBA_ANYFLAGS_RELEASE)
 		return(CORBA_TRUE);
 	else
 		return(CORBA_FALSE);
@@ -110,10 +110,26 @@ CORBA_boolean CORBA_sequence_get_release(void *seq)
  * freeable by CORBA_free(), so we can't use memory chunks here in any reasonable
  * fashion.
  */
+gpointer
+CORBA_any__free(gpointer mem, gpointer func_data, CORBA_boolean free_strings)
+{
+	CORBA_any *aval = mem;
+
+	if(aval->_release)
+		ORBit_free(aval->_value, free_strings);
+	CORBA_Object_release((CORBA_Object)aval->_type, NULL);
+
+	return aval + 1;
+}
 
 CORBA_any *CORBA_any_alloc(void)
 {
-	return ORBit_alloc(sizeof(CORBA_any), NULL, NULL);
+	CORBA_any *retval = ORBit_alloc(sizeof(CORBA_any), &CORBA_any__free,
+					GINT_TO_POINTER(1));
+
+	memset(retval, 0, sizeof(CORBA_any)); /* Make things easier on stubs */
+
+	return retval;
 }
 
 /*
@@ -124,34 +140,48 @@ CORBA_boolean ORBit_any_equivalent(CORBA_any obj, CORBA_any any, CORBA_Environme
 	return(CORBA_FALSE);
 }
 
+/* This is needed by skels, that generate a __free function when they see
+   the TypeCode interface */
+gpointer
+CORBA_TypeCode__free(gpointer mem, gpointer func_data, CORBA_boolean free_strings)
+{
+	CORBA_Object_release(*(CORBA_Object *)mem, NULL);
+	return ((char *)mem) + sizeof(CORBA_TypeCode);
+}
+
 CORBA_char *CORBA_string_dup(CORBA_char *string)
 {
-	CORBA_char *retval;
-
 	if(!string)
 		return NULL;
 
-	retval = ORBit_alloc(strlen(string)+1, NULL, NULL);
-
-	strcpy(retval, string);
-
-	return retval;
+	return strcpy(ORBit_alloc(strlen(string)+1, NULL, NULL), string);
 }
 
 CORBA_char *CORBA_string_alloc(CORBA_unsigned_long len)
 {
-	if(len)
-		return ORBit_alloc(len + 1, NULL, NULL);
-	else
-		return NULL;
+	return ORBit_alloc(len + 1, NULL, NULL);
 }
 
 CORBA_wchar *CORBA_wstring_alloc(CORBA_unsigned_long len)
 {
-	if(len)
-		return ORBit_alloc(len + 1, NULL, NULL);
-	else
-		return NULL;
+	return ORBit_alloc(len + 1, NULL, NULL);
+}
+
+gpointer
+CORBA_string__free(gpointer str, gpointer dat, CORBA_boolean free_strings)
+{
+	if(free_strings)
+		CORBA_free(*((gpointer *)str));
+	return str + sizeof(CORBA_char *);
+}
+
+gpointer CORBA_Object__free(gpointer str, gpointer dat, CORBA_boolean free_strings)
+{
+	CORBA_Environment ev;
+	CORBA_exception_init(&ev);
+	CORBA_Object_release(*((gpointer *)str), &ev);
+	CORBA_exception_free(&ev);
+	return str + sizeof(CORBA_Object);
 }
 
 /* 19.14 */
@@ -268,124 +298,80 @@ CORBA_fixed_d_s *CORBA_fixed_alloc(CORBA_unsigned_short d)
 
 void CORBA_free(void *storage)
 {
-	ORBit_free(storage);
+	ORBit_free(storage, CORBA_TRUE);
 }
 
-int ORBit_parse_inet(CORBA_Object obj, char *hostname, unsigned short port)
+int ORBit_parse_unixsock(CORBA_Object obj,
+			 char *sockpath,
+			 gboolean existing_only)
 {
-	obj->connection = GIOP_CONNECTION(iiop_connection_get(hostname, port));
-
-	if(obj->connection) {
-		giop_connection_ref(obj->connection);
-		return 0;
-	} else
+	if(!sockpath || !*sockpath)
 		return -1;
+	
+	obj->connection =
+		GIOP_CONNECTION(iiop_connection_unix_get(sockpath,
+							 existing_only));
+
+	if(!obj->connection)
+		return -1;
+
+	giop_connection_ref(obj->connection);
+	return 0;
 }
 
-/**************************** UNUSED ******************************/
-#if 0
-ORBitClassInfo **
-gen_parent_class_list(glong *parent_class_ids)
+int ORBit_parse_inet(CORBA_Object obj, char *hostname, unsigned short port,
+			 gboolean existing_only)
 {
-	int i, n;
-	ORBitClassInfo **retval;
+	obj->connection = GIOP_CONNECTION(iiop_connection_get(hostname, port, existing_only));
 
-	for(i = 0; parent_class_ids[i] >= 0; i++)
-		/* */ ;
-	n = i;
+	if(!obj->connection)
+		return -1;
+	giop_connection_ref(obj->connection);
+	return 0;
+}
 
-	retval = g_new(ORBitClassInfo *, n + 1);
+static const CORBA_unsigned_long zero_int = 0;
+struct iovec ORBit_default_principal_iovec = {(gpointer)&zero_int, sizeof(zero_int)};
 
-	for(i = 0; i < n; i++)
-		retval[i] = g_hash_table_lookup(orbit_class_byid,
-						(gpointer)parent_class_ids[i]);
-	retval[i] = NULL;
+void ORBit_set_default_principal(CORBA_Principal *principal)
+{
+	gpointer t;
+
+	if((gpointer)ORBit_default_principal_iovec.iov_base != (gpointer)&zero_int)
+		g_free(ORBit_default_principal_iovec.iov_base);
+
+	ORBit_default_principal_iovec.iov_len = principal->_length
+		+ sizeof(CORBA_unsigned_long);
+
+	t = ORBit_default_principal_iovec.iov_base =
+		g_malloc(ORBit_default_principal_iovec.iov_len);
+
+	memcpy(t, &principal->_length, sizeof(principal->_length));
+
+	t += sizeof(principal->_length);
+	memcpy(t, principal->_buffer, principal->_length);
+}
+
+CORBA_unsigned_long ORBit_class_assignment_counter = 0;
+
+/* XXX not thread-safe */
+CORBA_unsigned_long
+ORBit_register_class(PortableServer_ClassInfo *class_info)
+{
+	static GHashTable *class_assignments = NULL;
+	CORBA_unsigned_long retval;
+
+	if(!class_assignments)
+		class_assignments = g_hash_table_new(g_str_hash, g_str_equal);
+
+	/* This needs to be pre-increment - we don't want to give out
+	   classid 0, because (a) that is reserved for the base Object class
+	   (b) all the routines allocate a new id if the variable
+	   storing their ID == 0 */
+	retval = ++ORBit_class_assignment_counter;
+
+	g_hash_table_insert(class_assignments, (gpointer)class_info->class_name,
+			    GINT_TO_POINTER(retval));
 
 	return retval;
 }
-
-glong
-ORBit_register_class(char *name,
-		     gpointer method_stubs,
-		     gpointer method_skels,
-		     gpointer class_vtable_init_func,
-		     glong *parent_class_ids)
-{
-	ORBitClassInfo *newclass, **pclass;
-
-	if(!orbit_class_list) {
-		orbit_class_list = g_hash_table_new(g_str_hash, g_str_equal);
-		orbit_class_byid = g_hash_table_new(g_direct_hash, NULL);
-	}
-
-	newclass = g_hash_table_lookup(orbit_class_list, name);
-
-	if(!newclass) {
-		if(parent_class_ids) {
-			pclass = gen_parent_class_list(parent_class_ids);
-			g_return_val_if_fail(pclass != NULL, -1);
-		} else
-			pclass = NULL;
-
-		newclass = g_new(ORBitClassInfo, 1);
-      
-		newclass->name = g_strdup(name);
-		newclass->id = class_id_counter++;
-		newclass->method_stubs = method_stubs;
-		newclass->method_skels = method_skels;
-		newclass->class_vtable_init_func = class_vtable_init_func;
-		newclass->parent_classes = pclass;
-      
-		g_hash_table_insert(orbit_class_list,
-				    newclass->name, newclass);
-		g_hash_table_insert(orbit_class_byid,
-				    (gpointer)newclass->id, newclass);
-	}
-
-	return newclass->id;
-}
-
-static void
-init_object_and_recurse(CORBA_Object o,
-			ORBitClassInfo *class,
-			gboolean is_impl)
-{
-	int i;
-	gpointer *ptmp;
-
-	if(is_impl)
-		ptmp = class->method_skels;
-	else
-		ptmp = class->method_stubs;
-
-	for(i = 0; ptmp[i]; i++) /**/ ;
-
-	memcpy(o->vtable[class->id], ptmp, sizeof(gpointer) * i);
-
-	if(class->parent_classes) {
-		for(i = 0; class->parent_classes[i]; i++)
-			init_object_and_recurse(o, class->parent_classes[i],
-						is_impl);
-	}
-}
-
-void
-ORBit_initialize_object_vtable(CORBA_Object _handle_to_be,
-			       glong class_id,
-			       gboolean is_impl)
-{
-	ORBitClassInfo *c;
-
-	g_return_if_fail(class_id >= 0);
-	g_return_if_fail(_handle_to_be != NULL);
-
-	c = g_hash_table_lookup(orbit_class_byid, (gpointer) class_id);
-
-	g_return_if_fail(c != NULL);
-
-	_handle_to_be->vtable = g_new0(gpointer, class_id_counter);
-
-	init_object_and_recurse(_handle_to_be, c, is_impl);
-}
-#endif
-/**************************** END UNUSED ******************************/
