@@ -190,12 +190,14 @@ ORBit_create_object_with_info(GSList *profiles,
 	struct CORBA_Object_struct refcheck;
 
 	if(!type_id || !*type_id) {
+		g_warning("Failing object creation because object has no type");
 		CORBA_exception_set_system(ev, ex_CORBA_MARSHAL,
 					   CORBA_COMPLETED_MAYBE);
 		return CORBA_OBJECT_NIL;
 	}
 
 	if(g_slist_length(profiles) < 1) {
+		g_warning("Failing object creation because object has no profiles");
 		CORBA_exception_set_system(ev, ex_CORBA_MARSHAL,
 					   CORBA_COMPLETED_MAYBE);
 		return CORBA_OBJECT_NIL;
@@ -320,6 +322,9 @@ ORBit_demarshal_profile(GIOPRecvBuffer *recv_buffer, IOP_ProfileId profile_id)
 		break;
 	}
 error_exit:
+	g_message("demarshal_profile(): IIOP major is %d",
+		  object_info->iiop_major);
+	g_free(object_info);
 	if(codec!=NULL)
 		CDR_codec_free(codec);
 
@@ -398,6 +403,7 @@ ORBit_demarshal_object(GIOPRecvBuffer *recv_buffer, CORBA_ORB orb)
 		GET_ATOM(profile_id);
 		object_info=ORBit_demarshal_profile(recv_buffer, profile_id);
 		if(object_info==NULL) {
+			g_warning("Doing an error exit because demarshal_profile failed");
 			goto error_exit;
 		} else {
 			profiles=g_slist_append(profiles, object_info);
@@ -441,7 +447,8 @@ static void ORBit_marshal_profile(gpointer item, gpointer data)
 		codec->readonly = CORBA_FALSE;
 		codec->data_endian = codec->host_endian = FLAG_ENDIANNESS;
 		CDR_put_octet(codec, oc_endian);
-		CDR_put_octets(codec, (gpointer)iiopversion, sizeof(iiopversion));
+		CDR_put_octet(codec, iiopversion[0]);
+		CDR_put_octet(codec, iiopversion[1]);
 		CDR_put_string(codec, info->tag.iopinfo.host);
 		CDR_put_ushort(codec, info->tag.iopinfo.port);
 		CDR_put_ulong(codec, info->object_key._length);
