@@ -171,40 +171,44 @@ ORBit_POA_set_policy(PortableServer_POA poa,
 		poa->id_assignment = ((PortableServer_IdAssignmentPolicy)policy)->value;
 		break;
 	case PortableServer_IMPLICIT_ACTIVATION_POLICY_ID:
-		if(((PortableServer_ImplicitActivationPolicy)policy)->value
-		   == PortableServer_IMPLICIT_ACTIVATION
-		   && !((poa->id_assignment ==
-			 PortableServer_SYSTEM_ID)
-			&& (poa->servant_retention ==
-			    PortableServer_RETAIN))) {
-			CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
-					    ex_PortableServer_POA_InvalidPolicy,
-					    NULL);
-		} else
-			poa->implicit_activation = ((PortableServer_ImplicitActivationPolicy)policy)->value;
+	  poa->implicit_activation = ((PortableServer_ImplicitActivationPolicy)policy)->value;
 		break;
 	case PortableServer_SERVANT_RETENTION_POLICY_ID:
-		if(((PortableServer_ServantRetentionPolicy)policy)->value
-		   == PortableServer_NON_RETAIN
-		   && !((poa->request_processing ==
-			 PortableServer_USE_DEFAULT_SERVANT)
-			|| (poa->request_processing ==
-			    PortableServer_USE_SERVANT_MANAGER))) {
-
-			CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
-					    ex_PortableServer_POA_InvalidPolicy,
-					    NULL);
-		} else
-			poa->servant_retention = ((PortableServer_ServantRetentionPolicy)policy)->value;
+	  poa->servant_retention = ((PortableServer_ServantRetentionPolicy)policy)->value;
 		break;
 	case PortableServer_REQUEST_PROCESSING_POLICY_ID:
-		poa->servant_retention = ((PortableServer_ServantRetentionPolicy)policy)->value;
+		poa->request_processing = ((PortableServer_ServantRetentionPolicy)policy)->value;
 		break;
 	default:
 		g_warning("Unknown policy type, cannot set it on this POA");
 	}
 }
 
+
+static void
+ORBit_POA_check_policy_conflicts(PortableServer_POA poa,
+		     CORBA_Environment *ev)
+{
+
+  /* Check for those policy combinations that aren't allowed */
+  if ((poa->servant_retention == PortableServer_NON_RETAIN &&
+       poa->request_processing == PortableServer_USE_ACTIVE_OBJECT_MAP_ONLY) ||
+      
+      (poa->request_processing == PortableServer_USE_DEFAULT_SERVANT &&
+       poa->id_uniqueness == PortableServer_UNIQUE_ID) ||
+      
+      (poa->implicit_activation == PortableServer_IMPLICIT_ACTIVATION &&
+       (poa->id_assignment == PortableServer_USER_ID ||
+	poa->servant_retention == PortableServer_NON_RETAIN))
+      )
+    {
+      CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
+			  ex_PortableServer_POA_InvalidPolicy,
+			  NULL);
+    }
+}
+      
+
 static void
 ORBit_POA_set_policylist(PortableServer_POA poa,
 			 CORBA_PolicyList *policies,
@@ -243,33 +247,19 @@ ORBit_POA_new(CORBA_ORB orb,
 
 	if(ev->_major != CORBA_NO_EXCEPTION) goto error;
 
+	/* If no POAManager was specified, create one */
+	if(the_POAManager == NULL) {
+	  the_POAManager = ORBit_POAManager_new(ev);
+	}	  
+
 	/* Register this poa with the poa manager */
 	if(the_POAManager != NULL)
 		ORBit_POAManager_register_poa(the_POAManager,poa,ev);
 	if(ev->_major != CORBA_NO_EXCEPTION) goto error;
 
-	g_return_val_if_fail(ev->_major == CORBA_NO_EXCEPTION, NULL);
-
-	/* Need to initialise poa policies etc.. here */
-	poa->thread = PortableServer_ORB_CTRL_MODEL;
-	poa->lifespan = PortableServer_TRANSIENT;
-	poa->id_uniqueness = PortableServer_UNIQUE_ID;
-	poa->id_assignment = PortableServer_SYSTEM_ID;
-	poa->servant_retention = PortableServer_RETAIN;
-	poa->request_processing = PortableServer_USE_ACTIVE_OBJECT_MAP_ONLY;
-	poa->implicit_activation = PortableServer_NO_IMPLICIT_ACTIVATION;
-	ORBit_POA_set_policylist(poa, policies, ev);
-	if(ev->_major != CORBA_NO_EXCEPTION) goto error;
-
-	/* copy the name  */
-	poa->the_name = CORBA_string_dup(adapter_name);
-
 	/* Wire up the poa_manager */
 	poa->the_POAManager = the_POAManager;
 
-	poa->active_object_map = g_hash_table_new((GHashFunc)g_sequence_octet_hash,
-						  (GCompareFunc)g_sequence_octet_compare);
-
 	/* Initialise the child poas table */
 	poa->child_POAs = NULL;      /* initialise the slist */
 
@@ -281,6 +271,29 @@ ORBit_POA_new(CORBA_ORB orb,
 
 	poa->orb = orb;
 
+	g_return_val_if_fail(ev->_major == CORBA_NO_EXCEPTION, NULL);
+
+	/* Need to initialise poa policies etc.. here */
+	poa->thread = PortableServer_ORB_CTRL_MODEL;
+	poa->lifespan = PortableServer_TRANSIENT;
+	poa->id_uniqueness = PortableServer_UNIQUE_ID;
+	poa->id_assignment = PortableServer_SYSTEM_ID;
+	poa->servant_retention = PortableServer_RETAIN;
+	poa->request_processing = PortableServer_USE_ACTIVE_OBJECT_MAP_ONLY;
+	poa->implicit_activation = PortableServer_NO_IMPLICIT_ACTIVATION;
+	if (policies) {
+	  ORBit_POA_set_policylist(poa, policies, ev);
+	  ORBit_POA_check_policy_conflicts(poa, ev);
+	  if(ev->_major != CORBA_NO_EXCEPTION) goto error;
+	}
+
+	/* copy the name  */
+	poa->the_name = CORBA_string_dup(adapter_name);
+
+	poa->active_object_map = g_hash_table_new((GHashFunc)g_sequence_octet_hash,
+						  (GCompareFunc)g_sequence_octet_compare);
+
+
 	return poa;
 
 error:
