@@ -336,18 +336,30 @@ PortableServer_POA_create_POA
 		CORBA_Environment *ev)
 {
 	PortableServer_POA new_poa = NULL;
-
+	PortableServer_POA check_poa = NULL;
+	
 	if(policies)
 		g_warning("non-default POA policies aren't implemented yet");
 
-	new_poa = ORBit_POA_new(poa->orb,
-				adapter_name, a_POAManager, policies, ev);
-	new_poa->parent_poa = poa;
+	/* Check for a child POA by the same name in parent */
+	check_poa = PortableServer_POA_find_POA(poa,adapter_name,
+						FALSE, ev);
+	CORBA_exception_free (ev);
 
-	if(ev->_major == CORBA_NO_EXCEPTION)
-		ORBit_POA_add_child(poa, new_poa, ev);
+	if (!check_poa) {
+	  new_poa = ORBit_POA_new(poa->orb,
+				  adapter_name, a_POAManager, policies, ev);
+	} else {
+	  CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
+			  ex_PortableServer_POA_AdapterAlreadyExists,
+			  NULL);
+	  new_poa = NULL;
+	}
 
-	ev->_major = CORBA_NO_EXCEPTION;
+	if(ev->_major == CORBA_NO_EXCEPTION) {
+	  new_poa->parent_poa = poa;
+	  ORBit_POA_add_child(poa, new_poa, ev);
+	}
 
 	return new_poa;
 }
