@@ -24,6 +24,8 @@
 #include <string.h>
 #include <syslog.h>
 #include <signal.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 
 /*** App-specific servant structures ***/
 
@@ -259,7 +261,7 @@ static RegisteredName *rn_unref(RegisteredName *nom)
     return NULL;
 }
 
-static void do_nothing(void) { /* for alarm() handling */ }
+static void do_exit(void) { /* for alarm() handling */ _exit(1); }
 
 static gint
 server_is_alive(CORBA_Object server)
@@ -273,13 +275,14 @@ server_is_alive(CORBA_Object server)
   childpid = fork();
 
   if(!childpid) {
+    signal(SIGALRM, do_exit);
+    alarm(1);
     cnx = _ORBit_object_get_connection(server);
-    syslog(LOG_DEBUG, "server_is_alive: cnx = %p\n", cnx);
+    syslog(LOG_DEBUG, "server_is_alive: cnx[%s] = %p\n",
+	   server->object_id, cnx);
     _exit(cnx == 0);
   }
 
-  signal(SIGALRM, do_nothing);
-  alarm(3);
   itmp = waitpid(childpid, &exitstatus, 0);
 
   if(itmp < 0) return FALSE;
@@ -337,6 +340,8 @@ impl_CosNaming_NamingContext_bind(impl_POA_CosNaming_NamingContext * servant,
 {
   RegisteredName *nom;
 
+  if(n->_length < 1) goto raise_exc;
+
   nom = g_hash_table_lookup(servant->names, n->_buffer);
 
   if(n->_length == 1) {
@@ -362,6 +367,7 @@ impl_CosNaming_NamingContext_bind(impl_POA_CosNaming_NamingContext * servant,
   } else {
 
     CosNaming_NamingContext_NotFound *exdata;
+  raise_exc:
 
     exdata = CosNaming_NamingContext_NotFound__alloc();
     exdata->why = nom?CosNaming_NamingContext_not_context:CosNaming_NamingContext_missing_node;
@@ -380,6 +386,8 @@ impl_CosNaming_NamingContext_rebind(impl_POA_CosNaming_NamingContext * servant,
 {
   RegisteredName *nom;
 
+  if(n->_length < 1) goto raise_exc;
+
   nom = g_hash_table_lookup(servant->names, n->_buffer);
 
   if(n->_length == 1) {
@@ -398,8 +406,8 @@ impl_CosNaming_NamingContext_rebind(impl_POA_CosNaming_NamingContext * servant,
     n2._buffer++;
     CosNaming_NamingContext_rebind(nom->obj, &n2, obj, ev);
   } else {
-
     CosNaming_NamingContext_NotFound *exdata;
+  raise_exc:
 
     exdata = CosNaming_NamingContext_NotFound__alloc();
     exdata->why = nom?CosNaming_NamingContext_not_context:CosNaming_NamingContext_missing_node;
@@ -418,6 +426,8 @@ impl_CosNaming_NamingContext_bind_context(impl_POA_CosNaming_NamingContext * ser
 {
   RegisteredName *nom;
 
+  if(n->_length < 1) goto raise_exc;
+
   nom = g_hash_table_lookup(servant->names, n->_buffer);
 
   if(n->_length == 1) {
@@ -443,6 +453,7 @@ impl_CosNaming_NamingContext_bind_context(impl_POA_CosNaming_NamingContext * ser
   } else {
 
     CosNaming_NamingContext_NotFound *exdata;
+  raise_exc:
 
     exdata = CosNaming_NamingContext_NotFound__alloc();
     exdata->why = nom?CosNaming_NamingContext_not_context:CosNaming_NamingContext_missing_node;
@@ -461,6 +472,7 @@ impl_CosNaming_NamingContext_rebind_context(impl_POA_CosNaming_NamingContext * s
 {
   RegisteredName *nom;
 
+  if(n->_length < 1) goto raise_exc;
   nom = g_hash_table_lookup(servant->names, n->_buffer);
 
   if(n->_length == 1) {
@@ -479,8 +491,8 @@ impl_CosNaming_NamingContext_rebind_context(impl_POA_CosNaming_NamingContext * s
     n2._buffer++;
     CosNaming_NamingContext_rebind_context(nom->obj, &n2, nc, ev);
   } else {
-
     CosNaming_NamingContext_NotFound *exdata;
+  raise_exc:
 
     exdata = CosNaming_NamingContext_NotFound__alloc();
     exdata->why = nom?CosNaming_NamingContext_not_context:CosNaming_NamingContext_missing_node;
@@ -498,6 +510,7 @@ impl_CosNaming_NamingContext_resolve(impl_POA_CosNaming_NamingContext * servant,
 {
   RegisteredName *nom;
 
+  if(n->_length < 1) goto raise_exc;
   nom = g_hash_table_lookup(servant->names, n->_buffer);
   if(!nom) {
     CosNaming_NamingContext_NotFound *exdata;
@@ -513,6 +526,7 @@ impl_CosNaming_NamingContext_resolve(impl_POA_CosNaming_NamingContext * servant,
        return CORBA_Object_duplicate(nom->obj, ev);
     else {
       CosNaming_NamingContext_NotFound *exdata;
+    raise_exc:
 
       impl_CosNaming_NamingContext_unbind(servant, n, ev);
 
@@ -541,6 +555,7 @@ impl_CosNaming_NamingContext_unbind(impl_POA_CosNaming_NamingContext * servant,
 {
   RegisteredName *nom;
 
+  if(n->_length < 1) goto raise_exc;
   nom = g_hash_table_lookup(servant->names, n->_buffer);
 
   if(n->_length == 1 && nom) {
@@ -554,6 +569,7 @@ impl_CosNaming_NamingContext_unbind(impl_POA_CosNaming_NamingContext * servant,
     CosNaming_NamingContext_unbind(nom->obj, &n2, ev);
   } else {
     CosNaming_NamingContext_NotFound *exdata;
+  raise_exc:
 
     exdata = CosNaming_NamingContext_NotFound__alloc();
     exdata->why = nom?CosNaming_NamingContext_not_context:CosNaming_NamingContext_missing_node;
@@ -579,6 +595,7 @@ impl_CosNaming_NamingContext_bind_new_context(impl_POA_CosNaming_NamingContext *
   RegisteredName *nom;
   CosNaming_NamingContext retval;
 
+  if(n->_length < 1) goto raise_exc;
   nom = g_hash_table_lookup(servant->names, n->_buffer);
 
   if(n->_length == 1) {
@@ -604,6 +621,7 @@ impl_CosNaming_NamingContext_bind_new_context(impl_POA_CosNaming_NamingContext *
     return CosNaming_NamingContext_bind_new_context(nom->obj, &n2, ev);
   } else {
     CosNaming_NamingContext_NotFound *exdata;
+  raise_exc:
 
     exdata = CosNaming_NamingContext_NotFound__alloc();
     exdata->why = nom?CosNaming_NamingContext_not_context:CosNaming_NamingContext_missing_node;
